# coding=utf-8
from __future__ import absolute_import, division
from __future__ import print_function, unicode_literals

from itertools import chain, groupby
from os import path as osp
import threading

import numpy as np
from six.moves import range, cPickle as pickle, zip
from six import next

from smqtk.algorithms.nn_index import NearestNeighborsIndex
from smqtk.exceptions import ReadOnlyError
from smqtk.representation import get_descriptor_index_impls
from smqtk.representation.descriptor_element import elements_to_matrix
from smqtk.utils import plugin, merge_dict
from smqtk.utils.file_utils import safe_create_dir


CHUNK_SIZE = 5000


class MRPTNearestNeighborsIndex (NearestNeighborsIndex):
    """
    Nearest Neighbors index that uses the MRPT algorithm of [Hyvönen et
    al](https://arxiv.org/abs/1509.06957).

    Multiple Random Projection Trees (MRPT) combines multiple shallow binary
    trees of a set depth to quickly search for near neighbors. Each tree has a
    separate set of random projections used to divide the dataset into a tree
    structure. This algorithm differs from most RP tree implementations in
    that all separators at a particular level in the tree share the same
    projection, to save on space. Every branch partitions a set of points into
    two equal portions relative to the corresponding random projection.

    On query, the leaf corresponding to the query vector is found in each
    tree. The neighbors are drawn from the set of points that are in the most
    leaves.

    The performance will depend on settings for the parameters:

    - If `depth` is too high, then the leaves will not have enough points
        to satisfy a query, and num_trees will need to be higher in order to
        compensate. If `depth` is too low, then performance may suffer because
        the leaves are large. If `N` is the size of the dataset, and `L =
        N/2^{depth}`, then leaves should be small enough that all
        `num_trees*L` descriptors that result from a query will fit easily in
        cache. Since query complexity is linear in `depth`, this parameter
        should be kept as low as possible.
    - The `num_trees` parameter will lower the variance of the results for
        higher values, but at the cost of using more memory on any particular
        query. As a rule of thumb, for a given value of `k`, num_trees should
        be about `3k/L`.
    """

    @classmethod
    def is_usable(cls):
        return True

    @classmethod
    def get_default_config(cls):
        """
        Generate and return a default configuration dictionary for this class.
        This will be primarily used for generating what the configuration
        dictionary would look like for this class without instantiating it.

        By default, we observe what this class's constructor takes as
        arguments, turning those argument names into configuration dictionary
        keys. If any of those arguments have defaults, we will add those
        values into the configuration dictionary appropriately. The dictionary
        returned should only contain JSON compliant value types.

        It is not be guaranteed that the configuration dictionary returned
        from this method is valid for construction of an instance of this
        class.

        :return: Default configuration dictionary for the class.
        :rtype: dict

        """
        default = super(MRPTNearestNeighborsIndex, cls).get_default_config()

        di_default = plugin.make_config(get_descriptor_index_impls())
        default['descriptor_set'] = di_default

        return default

    @classmethod
    def from_config(cls, config_dict, merge_default=True):
        """
        Instantiate a new instance of this class given the configuration
        JSON-compliant dictionary encapsulating initialization arguments.

        This method should not be called via super unless and instance of the
        class is desired.

        :param config_dict: JSON compliant dictionary encapsulating
            a configuration.
        :type config_dict: dict

        :param merge_default: Merge the given configuration on top of the
            default provided by ``get_default_config``.
        :type merge_default: bool

        :return: Constructed instance from the provided config.
        :rtype: MRPTNearestNeighborsIndex

        """
        if merge_default:
            cfg = cls.get_default_config()
            merge_dict(cfg, config_dict)
        else:
            cfg = config_dict

        cfg['descriptor_set'] = \
            plugin.from_plugin_config(cfg['descriptor_set'],
                                      get_descriptor_index_impls())

        return super(MRPTNearestNeighborsIndex, cls).from_config(cfg, False)

    def __init__(self, descriptor_set, index_filepath=None,
                 parameters_filepath=None, read_only=False,
                 # Parameters for building an index
                 num_trees=10, depth=1, random_seed=None,
                 pickle_protocol=pickle.HIGHEST_PROTOCOL,
                 use_multiprocessing=False):
        """
        Initialize MRPT index properties. Does not contain a queryable index
        until one is built via the ``build_index`` method, or loaded from
        existing model files.

        :param descriptor_set: Index in which DescriptorElements will be
            stored.
        :type descriptor_set: smqtk.representation.DescriptorIndex

        :param index_filepath: Optional file location to load/store MRPT index
            when initialized and/or built.

            If not configured, no model files are written to or loaded from
            disk.
        :type index_filepath: None | str

        :param parameters_filepath: Optional file location to load/save index
            parameters determined at build time.

            If not configured, no model files are written to or loaded from
            disk.
        :type parameters_filepath: None | str

        :param read_only: If True, `build_index` will error if there is an
            existing index. False by default.
        :type read_only: bool

        :param num_trees: The number of trees that will be generated for the
            data structure
        :type num_trees: int

        :param depth: The depth of the trees
        :type depth: int

        :param random_seed: Integer to use as the random number generator
            seed.
        :type random_seed: int

        :param pickle_protocol: The protocol version to be used by the pickle
            module to serialize class information
        :type pickle_protocol: int

        :param use_multiprocessing: Whether or not to use discrete processes
            as the parallelization agent vs python threads.
        :type use_multiprocessing: bool

        """
        super(MRPTNearestNeighborsIndex, self).__init__()

        self._read_only = read_only
        self._use_multiprocessing = use_multiprocessing
        self._descriptor_set = descriptor_set
        self._pickle_protocol = pickle_protocol

        def normpath(p):
            return (p and osp.abspath(osp.expanduser(p))) or p

        # Lock for model component access.
        self._model_lock = threading.RLock()

        self._index_filepath = normpath(index_filepath)
        self._index_param_filepath = normpath(parameters_filepath)
        # Now they're either None or an absolute path

        # parameters for building an index
        if depth < 1:
            raise ValueError("The depth may not be less than 1.")
        self._depth = depth
        if num_trees < 1:
            raise ValueError("The number of trees must be positive.")
        self._num_trees = num_trees

        # Set the list of trees to an empty list to have a sane value
        self._trees = []

        #: :type: None | int
        self._rand_seed = None
        if random_seed:
            self._rand_seed = int(random_seed)

        # Load the index/parameters if one exists
        if self._has_model_files():
            self._log.debug("Found existing model files. Loading.")
            self._load_mrpt_model()

    def get_config(self):
        return {
            "descriptor_set": plugin.to_plugin_config(self._descriptor_set),
            "index_filepath": self._index_filepath,
            "parameters_filepath": self._index_param_filepath,
            "read_only": self._read_only,
            "random_seed": self._rand_seed,
            "pickle_protocol": self._pickle_protocol,
            "use_multiprocessing": self._use_multiprocessing,
            "depth": self._depth,
            "num_trees": self._num_trees,
        }

    def _has_model_files(self):
        """
        check if configured model files are configured and exist
        """
        return (self._index_filepath and
                osp.isfile(self._index_filepath) and
                self._index_param_filepath and
                osp.isfile(self._index_param_filepath))

    def _build_multiple_trees(self, chunk_size=CHUNK_SIZE):
        """
        Build an MRPT structure
        """
        sample = next(self._descriptor_set.iterdescriptors())
        sample_v = sample.vector()
        n = self.count()
        d = sample_v.size
        leaf_size = n / (1 << self._depth)

        self._log.debug(
            "Building %d trees (T) of depth %d (l) from %g descriptors (N) "
            "of length %g",
            self._num_trees, self._depth, n, d)
        self._log.debug(
            "Leaf size             (L = N/2^l)  ~ %g/2^%d = %g",
            n, self._depth, leaf_size)
        self._log.debug(
            "UUIDs stored                (T*N)  = %g * %g = %g",
            self._num_trees, n, self._num_trees*n)
        self._log.debug(
            "Examined UUIDs              (T*L)  ~ %g * %g = %g",
            self._num_trees, leaf_size, self._num_trees*leaf_size)
        self._log.debug(
            "Examined/DB size  (T*L/N = T/2^l)  ~ %g/%g = %.3f",
            self._num_trees*leaf_size, n, self._num_trees*leaf_size/n)

        if (1 << self._depth) > n:
            self._log.warn(
                "There are insufficient elements (%d < 2^%d) to populate "
                "all the leaves of the tree. Consider lowering the depth "
                "parameter.", n, self._depth)

        self._log.debug("Projecting onto random bases")
        # Build all the random bases and the projections at the same time
        # (_num_trees * _depth shouldn't really be that high -- if it is,
        # you're a monster)
        if self._rand_seed is not None:
            np.random.seed(self._rand_seed)
        random_bases = np.random.randn(self._num_trees, d, self._depth)
        projs = np.empty((n, self._num_trees, self._depth), dtype=np.float64)
        # Load the data in chunks (because n * d IS high)
        pts_array = np.empty((chunk_size, d), sample_v.dtype)
        # Enumerate the descriptors and div the index by the chunk size
        # (causes each loop to only deal with at most chunk_size descriptors at
        # a time).
        for k, g in groupby(enumerate(self._descriptor_set.iterdescriptors()),
                            lambda pair: pair[0] // chunk_size):
            # Items are still paired so extract the descriptors
            chunk = list(desc for (i, desc) in g)
            # Take care of dangling end piece
            k_beg = k * chunk_size
            k_end = min((k+1) * chunk_size, n)
            k_len = k_end - k_beg
            # Run the descriptors through elements_to_matrix
            elements_to_matrix(
                chunk, mat=pts_array, report_interval=1.0,
                use_multiprocessing=self._use_multiprocessing)
            # Insert into projection matrix
            projs[k_beg:k_end] = pts_array[:k_len].dot(random_bases)
        del pts_array

        self._log.debug("Constructing trees")
        desc_ids = list(self._descriptor_set.keys())
        # Start with no trees
        self._trees = []
        for t in range(self._num_trees):
            # Array of splits is a packed tree
            splits = np.empty(((1 << self._depth) - 1,), np.float64)

            self._log.debug("Constructing tree #%d", t+1)

            # Build the tree & store it
            leaves = self._build_single_tree(projs[:, t], splits)
            leaves = [[desc_ids[idx] for idx in leaf]
                      for leaf in leaves]
            self._trees.append({
                'random_basis': (random_bases[t]),
                'splits': splits,
                'leaves': leaves
            })

    def _build_single_tree(self, proj, splits):
        """
        Build a single RP tree for fast kNN search

        :param proj: Projections of the dataset for this tree
        :type proj: np.ndarray (N, levels)

        :param splits: (2^depth-1) array of splits corresponding to leaves
                       (tree, where immediate descendants follow parents;
                       index i's children are 2i+1 and 2i+2
        :type splits: np.ndarray

        :return: Tree of splits and list of index arrays for each leaf
        :rtype: list[np.ndarray]
        """
        def _build_recursive(indices, level=0, split_index=0):
            """
            Descend recursively into tree to build it, setting splits and
            returning indices for leaves

            :param indices: The current set of indices before partitioning
            :param level: The level in the tree
            :param split_index: The index of the split to set

            :return: A list of arrays representing leaf membership
            :rtype: list[np.ndarray]
            """
            # If we're at the bottom, no split, just return the set
            if level == self._depth:
                return [indices]

            n = indices.size
            # If we literally don't have enough to populate the leaf, make it
            # empty
            if n < 1:
                return []

            # Get the random projections for these indices at this level
            # NB: Recall that the projection matrix has shape (levels, N)
            level_proj = proj[indices, level]

            # Split at the median if even, put median in upper half if not
            n_split = n // 2
            if n % 2 == 0:
                part_indices = np.argpartition(
                    level_proj, (n_split - 1, n_split))
                split_val = level_proj[part_indices[n_split - 1]]
                split_val += level_proj[part_indices[n_split]]
                split_val /= 2.0
            else:
                part_indices = np.argpartition(level_proj, n_split)
                split_val = level_proj[part_indices[n_split]]

            splits[split_index] = split_val

            # part_indices is relative to this block of values, recover
            # main indices
            left_indices = indices[part_indices[:n_split]]
            right_indices = indices[part_indices[n_split:]]

            # Descend into each split and get sub-splits
            left_out = _build_recursive(left_indices, level=level + 1,
                                        split_index=2 * split_index + 1)
            right_out = _build_recursive(right_indices, level=level + 1,
                                         split_index=2 * split_index + 2)

            # Assemble index set
            left_out.extend(right_out)
            return left_out

        return _build_recursive(np.arange(proj.shape[0]))

    def _save_mrpt_model(self):
        self._log.debug("Caching index and parameters: %s, %s",
                        self._index_filepath, self._index_param_filepath)
        if self._index_filepath:
            self._log.debug("Caching index: %s", self._index_filepath)
            safe_create_dir(osp.dirname(self._index_filepath))
            # noinspection PyTypeChecker
            with open(self._index_filepath, "wb") as f:
                pickle.dump(self._trees, f, self._pickle_protocol)
        if self._index_param_filepath:
            self._log.debug("Caching index params: %s",
                            self._index_param_filepath)
            safe_create_dir(osp.dirname(self._index_param_filepath))
            params = {
                "read_only": self._read_only,
                "num_trees": self._num_trees,
                "depth": self._depth,
            }
            # noinspection PyTypeChecker
            with open(self._index_param_filepath, "w") as f:
                pickle.dump(params, f, self._pickle_protocol)

    def _load_mrpt_model(self):
        self._log.debug("Loading index and parameters: %s, %s",
                        self._index_filepath, self._index_param_filepath)
        if self._index_param_filepath:
            self._log.debug("Loading index params: %s",
                            self._index_param_filepath)
            with open(self._index_param_filepath) as f:
                params = pickle.load(f)
            self._read_only = params['read_only']
            self._num_trees = params['num_trees']
            self._depth = params['depth']

        # Load the index
        if self._index_filepath:
            self._log.debug("Loading index: %s", self._index_filepath)
            # noinspection PyTypeChecker
            with open(self._index_filepath, "rb") as f:
                self._trees = pickle.load(f)

    def count(self):
        """
        :return: Number of elements in this index.
        :rtype: int
        """
        # Descriptor-set should already handle concurrency.
        return len(self._descriptor_set)

    def _build_index(self, descriptors):
        """
        Internal method to be implemented by sub-classes to build the index with
        the given descriptor data elements.

        Subsequent calls to this method should rebuild the current index.  This
        method shall not add to the existing index nor raise an exception to as
        to protect the current index.

        :param descriptors: Iterable of descriptor elements to build index
            over.
        :type descriptors:
            collections.Iterable[smqtk.representation.DescriptorElement]

        """
        with self._model_lock:
            if self._read_only:
                raise ReadOnlyError("Cannot modify container attributes due to "
                                    "being in read-only mode.")

            self._log.info("Building new MRPT index")

            self._log.debug("Clearing and adding new descriptor elements")
            # NOTE: It may be the case for some DescriptorIndex implementations,
            # this clear may interfere with iteration when part of the input
            # iterator of descriptors was this index's previous descriptor-set,
            # as is the case with ``update_index``.
            self._descriptor_set.clear()
            self._descriptor_set.add_many_descriptors(descriptors)

            self._log.debug('Building MRPT index')
            self._build_multiple_trees()

            self._save_mrpt_model()

    def _update_index(self, descriptors):
        """
        Internal method to be implemented by sub-classes to additively update
        the current index with the one or more descriptor elements given.

        If no index exists yet, a new one should be created using the given
        descriptors.

        *NOTE:* This implementation fully rebuilds the index using the current
        index contents merged with the provided new descriptor elements.

        :param descriptors: Iterable of descriptor elements to add to this
            index.
        :type descriptors:
            collections.Iterable[smqtk.representation.DescriptorElement]

        """
        with self._model_lock:
            if self._read_only:
                raise ReadOnlyError("Cannot modify container attributes due "
                                    "to being in read-only mode.")
            self._log.debug("Updating index by rebuilding with union. ")
            self.build_index(chain(self._descriptor_set, descriptors))

    def _remove_from_index(self, uids):
        """
        Internal method to be implemented by sub-classes to partially remove
        descriptors from this index associated with the given UIDs.

        :param uids: Iterable of UIDs of descriptors to remove from this index.
        :type uids: collections.Iterable[collections.Hashable]

        :raises KeyError: One or more UIDs provided do not match any stored
            descriptors.

        """
        with self._model_lock:
            if self._read_only:
                raise ReadOnlyError("Cannot modify container attributes due "
                                    "to being in read-only mode.")
            self._descriptor_set.remove_many_descriptors(uids)
            self.build_index(self._descriptor_set)

    def _nn(self, d, n=1):
        """
        Internal method to be implemented by sub-classes to return the nearest
        `N` neighbors to the given descriptor element.

        When this internal method is called, we have already checked that there
        is a vector in ``d`` and our index is not empty.

        :param d: Descriptor element to compute the neighbors of.
        :type d: smqtk.representation.DescriptorElement

        :param n: Number of nearest neighbors to find.
        :type n: int

        :return: Tuple of nearest N DescriptorElement instances, and a tuple of
            the distance values to those neighbors.
        :rtype: (tuple[smqtk.representation.DescriptorElement], tuple[float])

        """
        def _query_single(tree):
            # Search a single tree for the leaf that matches the query
            # NB: random_basis has shape (levels, N)
            random_basis = tree['random_basis']
            proj_query = d.vector().dot(random_basis)
            splits = tree['splits']
            idx = 0
            for level in range(depth):
                split_point = splits[idx]
                # Look at the level'th coordinate of proj_query
                if proj_query[level] < split_point:
                    idx = 2 * idx + 1
                else:
                    idx = 2 * idx + 2

            # idx will be `2^depth - 1` greater than the position of the leaf
            # in the list
            idx -= ((1 << depth) - 1)
            return tree['leaves'][idx]

        def _exact_query(_uuids):
            set_size = len(_uuids)
            self._log.debug("Exact query requested with %d descriptors",
                            set_size)

            # Assemble the array to query from the descriptors that match
            d_v = d.vector()
            pts_array = np.empty((set_size, d_v.size), dtype=d_v.dtype)
            descriptors = self._descriptor_set.get_many_descriptors(_uuids)
            for i, desc in enumerate(descriptors):
                pts_array[i, :] = desc.vector()

            dists = ((pts_array - d_v) ** 2).sum(axis=1)

            if n > dists.shape[0]:
                self._log.warning(
                    "There were fewer descriptors (%d) in the set than "
                    "requested in the query (%d). Returning entire set.",
                    dists.shape[0], n)
            if n >= dists.shape[0]:
                return _uuids, dists

            near_indices = np.argpartition(dists, n - 1)[:n]
            return ([_uuids[idx] for idx in near_indices],
                    dists[near_indices])

        with self._model_lock:
            self._log.debug("Received query for %d nearest neighbors", n)

            depth, ntrees, db_size = self._depth, self._num_trees, self.count()
            leaf_size = db_size//(1 << depth)
            if leaf_size * ntrees < n:
                self._log.warning(
                    "The number of descriptors in a leaf (%d) times the "
                    "number of trees (%d) is less than the number of "
                    "descriptors requested by the query (%d). The query "
                    "result will be deficient.", leaf_size, ntrees, n)

            # Take union of all tree hits
            tree_hits = set()
            for t in self._trees:
                tree_hits.update(_query_single(t))

            hit_union = len(tree_hits)
            self._log.debug(
                "Query (k): %g, Hit union (h): %g, DB (N): %g, "
                "Leaf size (L = N/2^l): %g, Examined (T*L): %g",
                n, hit_union, db_size, leaf_size, leaf_size * ntrees)
            self._log.debug("k/L     = %.3f", n / leaf_size)
            self._log.debug("h/N     = %.3f", hit_union / db_size)
            self._log.debug("h/L     = %.3f", hit_union / leaf_size)
            self._log.debug("h/(T*L) = %.3f", hit_union / (leaf_size * ntrees))

            uuids, distances = _exact_query(list(tree_hits))
            order = distances.argsort()
            uuids, distances = zip(
                *((uuids[oidx], distances[oidx]) for oidx in order))

            self._log.debug("Returning query result of size %g", len(uuids))

            return (tuple(self._descriptor_set.get_many_descriptors(uuids)),
                    tuple(distances))


NN_INDEX_CLASS = MRPTNearestNeighborsIndex



import abc

from ._interface_classifier import Classifier


class SupervisedClassifier (Classifier):
    """
    Class of classifiers that are trainable via supervised training, i.e. are
    given specific descriptor examples for class labels.
    """

    @abc.abstractmethod
    def has_model(self):
        """
        :return: If this instance currently has a model loaded. If no model is
            present, classification of descriptors cannot happen (needs to be
            trained).
        :rtype: bool
        """

    def train(self, class_examples, **extra_params):
        """
        Train the supervised classifier model.

        If a model is already loaded, we will raise an exception in order to
        prevent accidental overwrite.

        If the same label is provided to both ``class_examples`` and ``kwds``,
        the examples given to the reference in ``kwds`` will prevail.

        :param class_examples: Dictionary mapping class labels to iterables of
            DescriptorElement training examples.
        :type class_examples: dict[collections.Hashable,
                 collections.Iterable[smqtk.representation.DescriptorElement]]

        :param extra_params: Dictionary with extra parameters for training.
        :type extra_params: dict[basestring, object]

        :raises ValueError: There were no class examples provided.
        :raises ValueError: Less than 2 classes were given.
        :raises RuntimeError: A model already exists in this instance.
            Following through with training would overwrite this model.
            Throwing an exception for information protection.

        """
        if self.has_model():
            raise RuntimeError("Instance currently has a model. Halting "
                               "training to prevent overwrite of existing "
                               "trained model.")

        if not class_examples:
            raise ValueError("No class examples were provided.")
        elif len(class_examples) < 2:
            raise ValueError("Need 2 or more classes for training. Given %d."
                             % len(class_examples))

        return self._train(class_examples, **extra_params)

    @abc.abstractmethod
    def _train(self, class_examples, **extra_params):
        """
        Internal method that trains the classifier implementation.

        This method is called after checking that there is not already a model
        trained, thus it can be assumed that no model currently exists.

        The class labels will have already been checked before entering this
        method, so it can be assumed that the ``class_examples`` will container
        at least two classes.

        :param class_examples: Dictionary mapping class labels to iterables of
            DescriptorElement training examples.
        :type class_examples: dict[collections.Hashable,
                 collections.Iterable[smqtk.representation.DescriptorElement]]

        :param extra_params: Dictionary with extra parameters for training.
        :type extra_params: None | dict[basestring, object]

        """



from ._interface import SmqtkAlgorithm

# Import module abstracts and plugin getter functions
from .classifier import Classifier, SupervisedClassifier, get_classifier_impls
from .descriptor_generator import DescriptorGenerator, \
    get_descriptor_generator_impls
from .nn_index import NearestNeighborsIndex, get_nn_index_impls
from .nn_index.hash_index import HashIndex, get_hash_index_impls
from .nn_index.lsh.functors import LshFunctor, get_lsh_functor_impls
from .relevancy_index import RelevancyIndex, get_relevancy_index_impls



from __future__ import division, print_function
import unittest

import mock
import numpy

from smqtk.algorithms.descriptor_generator import DescriptorGenerator
from smqtk.algorithms.descriptor_generator import get_descriptor_generator_impls
import smqtk.representation


class TestGetDescriptorGeneratorImpls (unittest.TestCase):

    def test_get_descriptors(self):
        m = get_descriptor_generator_impls()
        # Currently no types that are guaranteed available
        self.assertIsInstance(m, dict, "Should return a dictionary of "
                                       "class label-to-types")


class DummyDescriptorGenerator (DescriptorGenerator):
    """
    Shell implementation of abstract class in order to test abstract class
    functionality.
    """

    @classmethod
    def is_usable(cls):
        return True

    def get_config(self):
        return {}

    def valid_content_types(self):
        return {}

    def _compute_descriptor(self, data):
        return


class TestDescriptorGeneratorAbstract (unittest.TestCase):
    """
    Create mock object (look up mock module?) to test abstract super-class
    functionality in isolation.

    Test abstract super-class functionality where there is any
    """

    def test_compute_descriptor_invlaid_type(self):
        cd = DummyDescriptorGenerator()
        cd.valid_content_types = mock.Mock(return_value={'image/png'})

        mDataElement = mock.Mock(spec=smqtk.representation.DataElement)
        m_d = mDataElement()
        m_d.content_type.return_value = 'image/jpeg'

        mDescrElement = mock.Mock(spec=smqtk.representation.DescriptorElement)
        mDescriptorFactory = mock.Mock(
            spec=smqtk.representation.DescriptorElementFactory)
        m_factory = mDescriptorFactory()
        m_factory.new_descriptor.return_value = mDescrElement()

        self.assertRaises(ValueError, cd.compute_descriptor, m_d, m_factory)

    def test_computeDescriptor_validType_newVector(self):
        expected_image_type = 'image/png'
        expected_vector = numpy.random.randint(0, 100, 10)
        expected_uuid = "a unique ID"

        # Set up mock classes/responses
        cd = DummyDescriptorGenerator()
        cd.valid_content_types = mock.Mock(return_value={expected_image_type})
        cd._compute_descriptor = mock.Mock(return_value=expected_vector)

        mDataElement = mock.Mock(spec=smqtk.representation.DataElement)
        m_data = mDataElement()
        m_data.content_type.return_value = expected_image_type
        m_data.uuid.return_value = expected_uuid

        mDescrElement = mock.Mock(spec=smqtk.representation.DescriptorElement)
        mDescrElement().has_vector.return_value = False

        mDescriptorFactory = mock.Mock(
            spec=smqtk.representation.DescriptorElementFactory)
        m_factory = mDescriptorFactory()
        m_factory.new_descriptor.return_value = mDescrElement()

        # Call: matching content types, no existing descriptor for data
        d = cd.compute_descriptor(m_data, m_factory, overwrite=False)

        m_factory.new_descriptor.assert_called_once_with(cd.name, expected_uuid)
        self.assertTrue(cd._compute_descriptor.called)
        mDescrElement().set_vector.assert_called_once_with(expected_vector)
        self.assertIs(d, mDescrElement())

    def test_computeDescriptor_validType_existingVector(self):
        expected_image_type = 'image/png'
        expected_existing_vector = numpy.random.randint(0, 100, 10)
        expected_new_vector = numpy.random.randint(0, 100, 10)
        expected_uuid = "a unique ID"

        # Set up mock classes/responses
        mDataElement = mock.Mock(spec=smqtk.representation.DataElement)
        m_data = mDataElement()
        m_data.content_type.return_value = expected_image_type
        m_data.uuid.return_value = expected_uuid

        mDescrElement = mock.Mock(spec=smqtk.representation.DescriptorElement)
        mDescrElement().has_vector.return_value = True
        mDescrElement().vector.return_value = expected_existing_vector

        mDescriptorFactory = mock.Mock(
            spec=smqtk.representation.DescriptorElementFactory)
        m_factory = mDescriptorFactory()
        m_factory.new_descriptor.return_value = mDescrElement()

        cd = DummyDescriptorGenerator()
        cd.valid_content_types = mock.Mock(return_value={expected_image_type})
        cd._compute_descriptor = mock.Mock(return_value=expected_new_vector)

        # Call: matching content types, existing descriptor for data
        d = cd.compute_descriptor(m_data, m_factory, overwrite=False)

        self.assertTrue(mDescrElement().has_vector.called)
        self.assertFalse(cd._compute_descriptor.called)
        self.assertFalse(mDescrElement().set_vector.called)
        self.assertIs(d, mDescrElement())

    def test_computeDescriptor_validType_existingVector_overwrite(self):
        expected_image_type = 'image/png'
        expected_existing_vector = numpy.random.randint(0, 100, 10)
        expected_new_vector = numpy.random.randint(0, 100, 10)
        expected_uuid = "a unique ID"

        # Set up mock classes/responses
        mDataElement = mock.Mock(spec=smqtk.representation.DataElement)
        m_data = mDataElement()
        m_data.content_type.return_value = expected_image_type
        m_data.uuid.return_value = expected_uuid

        mDescrElement = mock.Mock(spec=smqtk.representation.DescriptorElement)
        mDescrElement().has_vector.return_value = True
        mDescrElement().vector.return_value = expected_existing_vector

        mDescriptorFactory = mock.Mock(
            spec=smqtk.representation.DescriptorElementFactory)
        m_factory = mDescriptorFactory()
        m_factory.new_descriptor.return_value = mDescrElement()

        cd = DummyDescriptorGenerator()
        cd.valid_content_types = mock.Mock(return_value={expected_image_type})
        cd._compute_descriptor = mock.Mock(return_value=expected_new_vector)

        # Call: matching content types, existing descriptor for data
        d = cd.compute_descriptor(m_data, m_factory, overwrite=True)

        self.assertFalse(mDescrElement().has_vector.called)
        self.assertTrue(cd._compute_descriptor.called)
        cd._compute_descriptor.assert_called_once_with(m_data)
        self.assertTrue(mDescrElement().set_vector.called)
        mDescrElement().set_vector.assert_called_once_with(expected_new_vector)
        self.assertIs(d, mDescrElement())

    def test_computeDescriptorAsync(self):
        # Only using threading because mock.Mock can't be serialized (pickled)
        # for multiprocessing IPC.

        # Mock input data
        m_d0 = mock.Mock(name='data-0',
                         spec=smqtk.representation.DataElement)()
        m_d1 = mock.Mock(name='data-1',
                         spec=smqtk.representation.DataElement)()

        m_d0.uuid.return_value = 'uuid-0'
        m_d1.uuid.return_value = 'uuid-1'

        m_factory = \
            mock.Mock(spec=smqtk.representation.DescriptorElementFactory)()

        def mock_compute(d, *_):
            if d is m_d0:
                return 1
            elif d is m_d1:
                return 2
            else:
                return None

        generator = DummyDescriptorGenerator()
        generator.compute_descriptor = mock.Mock(
            side_effect=mock_compute
        )

        m = generator.compute_descriptor_async([m_d0, m_d1], m_factory,
                                               overwrite=False, use_mp=False)

        self.assertEqual(len(m), 2)
        self.assertIn(m_d0.uuid(), m)
        self.assertIn(m_d1.uuid(), m)
        self.assertEqual(m[m_d0.uuid()], 1)
        self.assertEqual(m[m_d1.uuid()], 2)

        self.assertTrue(generator.compute_descriptor.called)
        self.assertEqual(generator.compute_descriptor.call_count, 2)
        generator.compute_descriptor.assert_any_call(m_d0, m_factory, False)
        generator.compute_descriptor.assert_any_call(m_d1, m_factory, False)

    def test_computeDescriptorAsync_failure(self):
        # Only using threading because mock.Mock can't be serialized (pickled)
        # for multiprocessing IPC.

        m_d0 = mock.Mock(spec=smqtk.representation.DataElement)()
        m_d1 = mock.Mock(spec=smqtk.representation.DataElement)()

        m_factory = \
            mock.Mock(spec=smqtk.representation.DescriptorElementFactory)()

        generator = DummyDescriptorGenerator()
        generator.compute_descriptor = mock.Mock(
            side_effect=RuntimeError("Intended exception")
        )

        self.assertRaises(
            RuntimeError,
            generator.compute_descriptor_async,
            [m_d0, m_d1], m_factory,
            procs=2,
            overwrite=False, use_mp=False
        )



from ._interface import SmqtkRepresentation

from .classification_element import ClassificationElement, \
    get_classification_element_impls
from .data_element import DataElement, get_data_element_impls
from .data_set import DataSet, get_data_set_impls
from .descriptor_element import DescriptorElement, get_descriptor_element_impls
from .descriptor_index import DescriptorIndex, get_descriptor_index_impls
from .key_value import KeyValueStore, get_key_value_store_impls

from .classification_element_factory import ClassificationElementFactory
from .descriptor_element_factory import DescriptorElementFactory



from __future__ import division, print_function
from smqtk.representation.data_set import get_data_set_impls


def test_plugin_getter():
    c = get_data_set_impls()
    assert isinstance(c, dict)
    # The following implementations at least should always be available.
    assert 'DataFileSet' in c
    assert 'DataMemorySet' in c



import threading

import six

from smqtk.exceptions import MissingLabelError
from smqtk.utils import Configurable, SmqtkObject, merge_dict, plugin

from . import get_classifier_impls
from ._defaults import DFLT_CLASSIFIER_FACTORY
from ._interface_classifier import Classifier


class ClassifierCollection (SmqtkObject, Configurable):
    """
    A collection of descriptively-labeled classifier instances for the purpose
    of applying all stored classifiers to one or more input descriptor
    elements.

    TODO: [optionally?] map a classification element factory per classifier.

    """

    EXAMPLE_KEY = '__example_label__'

    @classmethod
    def get_default_config(cls):
        c = super(ClassifierCollection, cls).get_default_config()

        # We list the label-classifier mapping on one level, so remove the
        # nested map parameter that can optionally be used in the constructor.
        del c['classifiers']

        # Add slot of a list of classifier plugin specifications
        c[cls.EXAMPLE_KEY] = plugin.make_config(get_classifier_impls())

        return c

    @classmethod
    def from_config(cls, config_dict, merge_default=True):
        """
        Instantiate a new instance of this class given the configuration
        JSON-compliant dictionary encapsulating initialization arguments.

        This method should not be called via super unless an instance of the
        class is desired.

        :param config_dict: JSON compliant dictionary encapsulating
            a configuration.
        :type config_dict: dict

        :param merge_default: Merge the given configuration on top of the
            default provided by ``get_default_config``.
        :type merge_default: bool

        :return: Constructed instance from the provided config.
        :rtype: ClassifierCollection

        """
        if merge_default:
            config_dict = merge_dict(cls.get_default_config(), config_dict)

        classifier_map = {}

        # Copying list of keys so we can update the dictionary as we loop.
        for label in list(config_dict.keys()):
            # Skip the example section.
            if label == cls.EXAMPLE_KEY:
                continue

            classifier_config = config_dict[label]
            classifier = plugin.from_plugin_config(classifier_config,
                                                   get_classifier_impls())
            classifier_map[label] = classifier

        # Don't merge back in "example" default
        return super(ClassifierCollection, cls).from_config(
            {'classifiers': classifier_map},
            merge_default=False
        )

    def __init__(self, classifiers=None, **labeled_classifiers):
        """
        :param classifiers: Optional dictionary of semantic label keys and
            Classifier instance values.
        :type classifiers: dict[str, Classifier]

        :param labeled_classifiers: Key-word arguments may be provided where
            the key used is considered the semantic label of the provided
            Classifier instance.
        :type labeled_classifiers: Classifier

        """
        self._label_to_classifier_lock = threading.RLock()
        self._label_to_classifier = {}

        # Go though classifiers map and key-word arguments, check that values
        # are actually classifiers.
        if classifiers is not None:
            for label, classifier in six.iteritems(classifiers):
                if not isinstance(classifier, Classifier):
                    raise ValueError("Found a non-Classifier instance value "
                                     "for key '%s'" % label)
                self._label_to_classifier[label] = classifier

        for label, classifier in six.iteritems(labeled_classifiers):
            if not isinstance(classifier, Classifier):
                raise ValueError("Found a non-Classifier instance value "
                                 "for key '%s'" % label)
            elif label in self._label_to_classifier:
                raise ValueError("Duplicate classifier label '%s' provided "
                                 "in key-word arguments." % label)
            self._label_to_classifier[label] = classifier

    def __enter__(self):
        """
        :rtype: IqrSession
        """
        self._label_to_classifier_lock.acquire()
        return self

    # noinspection PyUnusedLocal
    def __exit__(self, exc_type, exc_val, exc_tb):
        self._label_to_classifier_lock.release()

    def get_config(self):
        with self._label_to_classifier_lock:
            c = dict((label, plugin.to_plugin_config(classifier))
                     for label, classifier
                     in six.iteritems(self._label_to_classifier))
        return c

    def size(self):
        with self._label_to_classifier_lock:
            return len(self._label_to_classifier)

    __len__ = size

    def labels(self):
        """
        :return: Set of labels for currently collected classifiers.
        :rtype: set[str]
        """
        with self._label_to_classifier_lock:
            return set(self._label_to_classifier.keys())

    def add_classifier(self, label, classifier):
        """
        Add a classifier instance with associated descriptive label to this
        collection.

        :param label: String descriptive label for the classifier.
        :type label: str

        :param classifier: Classifier instance to collect.
        :type classifier: Classifier

        :raises ValueError: Classifier provided is not actually a classifier
            instance, or if the label provided already exists in this
            collection.

        :return: Self.
        :rtype: self

        """
        if not isinstance(classifier, Classifier):
            raise ValueError("Not given a Classifier instance (given type"
                             " %s)." % type(classifier))
        with self._label_to_classifier_lock:
            if label in self._label_to_classifier:
                raise ValueError("Duplicate label provided: '%s'" % label)
            self._label_to_classifier[label] = classifier
        return self

    def get_classifier(self, label):
        """
        Get the classifier instance for a given label.

        :param label: Label of the classifier to get.
        :type label: str

        :raises KeyError: No classifier for the given label.

        :return: Classifier instance.
        :rtype: Classifier

        """
        with self._label_to_classifier_lock:
            return self._label_to_classifier[label]

    def remove_classifier(self, label):
        """
        Remove a label-classifier pair from this collection.

        :param label: Label of the classifier to remove.
        :type label: str

        :raises KeyError: The given label does not reference a classifier in
            this collection.

        :return: Self.
        :rtype: self

        """
        with self._label_to_classifier_lock:
            del self._label_to_classifier[label]

    def classify(self, descriptor, labels=None,
                 factory=DFLT_CLASSIFIER_FACTORY, overwrite=False):
        """
        Apply all stored classifiers to the given descriptor element.

        We return a dictionary mapping the label of a stored classifier to the
        classifier element result produced by that classifier via the
        provided classification element factory.

        :param descriptor: Descriptor element to classify.
        :type descriptor: smqtk.representation.DescriptorElement

        :param labels: One or more labels of stored classifiers to use for
            classifying the given descriptor.  If None, use all stored
            classifiers.
        :type labels: Iterable[str]

        :param factory: Classification element factory.
        :type factory: ClassificationElementFactory

        :param overwrite: Force re-computation of the classification of the
            input descriptor.
        :type overwrite: bool

        :raises smqtk.exceptions.MissingLabelError: Some or all of the
            requested labels are missing.

        :return: Result dictionary of classifier labels to classification
            elements.
        :rtype: dict[str, smqtk.representation.ClassificationElement]

        """

        d_classifications = {}
        with self._label_to_classifier_lock:
            # TODO(paul.tunison): Parallelize?
            if labels is not None:
                # If we're missing some of the requested labels, complain
                missing_labels = set(labels) - self.labels()
                if missing_labels:
                    raise MissingLabelError(missing_labels)

                for label in labels:
                    classifier = self._label_to_classifier[label]
                    d_classifications[label] = classifier.classify(
                        descriptor, factory=factory, overwrite=overwrite)
            else:
                for label, classifier in six.iteritems(
                        self._label_to_classifier):
                    d_classifications[label] = classifier.classify(
                        descriptor, factory=factory, overwrite=overwrite)
        return d_classifications

    # TODO(paul.tunison): Classify many descriptors method when the need
    #   arises.



"""
LICENCE
-------
Copyright 2013-2015 by Kitware, Inc. All Rights Reserved. Please refer to
KITWARE_LICENSE.TXT for licensing information, or contact General Counsel,
Kitware, Inc., 28 Corporate Drive, Clifton Park, NY 12065.
"""

import logging
import multiprocessing
import time

from smqtk.utils.distance_kernel import DistanceKernel
from smqtk.utils.feature_memory import FeatureMemory


class TimedCache (object):
    """
    Be able to generally store FeatureMemory/DistanceKernel objects with timed
    expiry / deletion.
    """

    def __init__(self):
        self._lock = multiprocessing.RLock()

        # Mapping of a key to an object
        #: :type: dict[str, object]
        self._obj_cache = {}

        # Mapping of a key to the last access time for that object (UNIX time)
        #: :type: dict[str, float]
        self._obj_last_access = {}

        # Mapping of timeout values for each entry
        self._obj_timeouts = {}

    def __enter__(self):
        self._lock.acquire()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._lock.release()

    @property
    def _log(self):
        return logging.getLogger('.'.join([self.__module__,
                                           self.__class__.__name__]))

    def _check_expire(self):
        """
        Check all stored objects to see what has expired. Expired items are
        removed from our mappings.

        ASSUMING write lock has already been internally acquired.

        """
        self._log.debug("Checking entry expiration...")
        current_time = time.time()
        for key in self._obj_cache.keys():
            self._log.debug('  -> %s (type = %s)',
                            key, type(self._obj_cache[key]))
            # Remove if the key has a timeout, and the timeout period has been
            # exceeded (last access + timeout period <= current_time).
            if self._obj_timeouts[key] > 0 \
                    and current_time >= (self._obj_last_access[key]
                                         + self._obj_timeouts[key]):
                self._log.debug('     EXPIRED -- removing')
                # delete
                del self._obj_cache[key]
                del self._obj_last_access[key]
                del self._obj_timeouts[key]

    def get_rlock(self):
        return self._lock

    def keys(self):
        with self._lock:
            self._check_expire()
            return self._obj_cache.keys()

    def store(self, key, obj, timeout=0):
        """
        Store an object with a key. A timeout may be provided (in seconds).
        Defaults to no expiry.
        """
        with self._lock:
            self._check_expire()

            self._log.debug("storing '%s' (type=%s) with TO=%f",
                            key, type(obj), timeout)
            self._obj_cache[key] = obj
            self._obj_last_access[key] = time.time()
            self._obj_timeouts[key] = timeout

    def store_FeatureMemory(self, key, id_vector_file, bg_flags_file,
                            feature_mat_file, kernel_mat_file, timeout=0):
        fm = FeatureMemory.construct_from_files(id_vector_file,
                                                bg_flags_file,
                                                feature_mat_file,
                                                kernel_mat_file)
        self.store(key, fm, timeout)

    def store_DistanceKernel_symmetric(self, key, id_vector_file,
                                       kernel_mat_file, bg_flags_file=None,
                                       timeout=0):
        dk = DistanceKernel.construct_symmetric_from_files(id_vector_file,
                                                           kernel_mat_file,
                                                           bg_flags_file)
        self.store(key, dk, timeout)

    def store_DistanceKernel_asymmeric(self, key, row_ids_file, col_ids_file,
                                       kernel_mat_file, timeout=0):
        dk = DistanceKernel.construct_asymmetric_from_files(row_ids_file,
                                                            col_ids_file,
                                                            kernel_mat_file)
        self.store(key, dk, timeout)

    def get(self, key):
        """
        General getter function. When used with ProxyManager, I think this
        returns a copy of the remotely stored data.
        """
        with self._lock:
            self._check_expire()

            obj = self._obj_cache[key]
            self._log.debug("getting object '%s' (type=%s). "
                            "updating access time.",
                            key, type(obj))
            self._obj_last_access[key] = time.time()
            return obj

    def get_fm(self, key):
        obj = self.get(key)
        if isinstance(obj, FeatureMemory):
            return obj
        else:
            raise ValueError("Key not associated to a FeatureMemory object")

    def get_dk(self, key):
        obj = self.get(key)
        if isinstance(obj, DistanceKernel):
            return obj
        else:
            raise ValueError("Key not associated to a DistanceKernel object")

    def remove(self, key):
        """
        Remove a cached element if it exists. If it doesn't this is a no-op.
        """
        with self._lock:
            self._check_expire()

            if key in self._obj_cache:
                self._log.debug("removing entry '%s' (type=%s)",
                                key, type(self._obj_cache[key]))
                del self._obj_cache[key]
                del self._obj_last_access[key]
                del self._obj_timeouts[key]



"""
Tests for high level ``from_uri`` function, separate from the
``DataElement.from_uri`` class method.
"""
import unittest

import smqtk.exceptions
import smqtk.representation.data_element


# noinspection PyClassHasNoInit
class UnresolvableElement (smqtk.representation.data_element.DataElement):
    """ Does not implement from_uri, declaring no support for URI resolution """

    @classmethod
    def is_usable(cls):
        return True

    def __repr__(self):
        return super(UnresolvableElement, self).__repr__()

    def get_config(self):
        return {}

    def content_type(self):
        return None

    def is_empty(self):
        pass

    def get_bytes(self):
        return bytes()

    def set_bytes(self, b):
        pass

    def writable(self):
        pass


# noinspection PyClassHasNoInit
class ResolvableElement (smqtk.representation.data_element.DataElement):

    @classmethod
    def from_uri(cls, uri):
        """
        :type uri: str
        :rtype: ResolvableElement
        """
        if uri.startswith('resolvable://'):
            return ResolvableElement()

    @classmethod
    def is_usable(cls):
        return True

    def __repr__(self):
        return super(ResolvableElement, self).__repr__()

    def get_config(self):
        return {}

    def content_type(self):
        return None

    def is_empty(self):
        pass

    def get_bytes(self):
        return bytes()

    def set_bytes(self, b):
        pass

    def writable(self):
        pass


class TestDataElementHighLevelFromUri (unittest.TestCase):

    def test_no_classes(self):
        def impl_generator():
            return {}

        self.assertRaises(
            smqtk.exceptions.InvalidUriError,
            smqtk.representation.data_element.from_uri,
            'whatever',
            impl_generator
        )

    def test_no_resolvable_options(self):
        """
        when no DataElement implementations provide an implementation for
        the ``from_uri`` class method
        """
        def impl_generator():
            return {
                "ur1": UnresolvableElement,
                "ur2": UnresolvableElement,
            }

        self.assertRaises(
            smqtk.exceptions.InvalidUriError,
            smqtk.representation.data_element.from_uri,
            'something',
            impl_generator
        )

    def test_one_resolvable_option(self):
        """
        When at least one plugin can resolve a URI
        """
        def impl_generator():
            return {
                "e1": UnresolvableElement,
                "e2": ResolvableElement,
            }

        # URI that can be resolved by ResolvableElement
        self.assertIsInstance(
            smqtk.representation.data_element.from_uri(
                "resolvable://data",
                impl_generator
            ),
            ResolvableElement
        )

        # bad URI even though something can resolve it
        self.assertRaises(
            smqtk.exceptions.InvalidUriError,
            smqtk.representation.data_element.from_uri,
            'not_resolvable', impl_generator
        )



import unittest

from six.moves import cPickle

import smqtk.representation.classification_element.memory


class TestMemoryClassificationElement (unittest.TestCase):

    def test_serialization(self):
        e = smqtk.representation.classification_element.memory\
            .MemoryClassificationElement('test', 0)

        e2 = cPickle.loads(cPickle.dumps(e))
        self.assertEqual(e, e2)

        e.set_classification(a=0, b=1)
        e2 = cPickle.loads(cPickle.dumps(e))
        self.assertEqual(e, e2)



"""
LICENCE
-------
Copyright 2013 by Kitware, Inc. All Rights Reserved. Please refer to
LICENSE for licensing information, or contact General Counsel,
Kitware, Inc., 28 Corporate Drive, Clifton Park, NY 12065.

"""
# pyflann is a toolbox for approximate knn computation
# http://people.cs.ubc.ca/~mariusm/index.php/FLANN/FLANN
# Downloaded version can be founded in './external' directory

import bz2
import gzip
import logging
import os
import os.path as osp
import numpy as np
import csv


# These functions should only be called by colordescriptor.py module, which is
# only available if pyflann is available by the following check
try:
    import pyflann
except ImportError:
    pyflann = None


def flann_load_codebook(filename, is_rowwise=True):
    """
    Load codebook for quantization, and make it ready to be used for pyflann

    To build codebook for colorDescriptor, see the exampleCreateCodebook.py
    script that comes with the colorDescriptor ZIP package. This uses scipy.vq
    kmeans to build the centroid list given a set of colorDescriptor outputs.

    @param filename: file path to the codebook file
    @param is_rowwise: asks if codebook is stored as rowwise. If not, it will be
        transposed internally after loading.
    @return: row-wise numpy codebook
    """

    cbook = np.loadtxt(filename, delimiter=',')
    if not is_rowwise:
        cbook = cbook.T

    return cbook


def flann_build(codebook, flann_index_file=None, target_precision=0.99,
                log_level="info"):
    """
    Prepares quantization module based on pyflann

    @param codebook: loaded codebook in numpy format (row-wise)
    @param flann_index_file: path to the index file to read from.
    @param target_precision: amount of exactness of approximation, the lower,
        the faster
    @param log_level: log data format to be generatede by flann
    """
    log = logging.getLogger(__name__)

    flann = pyflann.FLANN()
    if not flann_index_file or not osp.isfile(flann_index_file):
        flann.build_index(codebook,
                          target_precision=target_precision,
                          log_level=log_level)

        if flann_index_file is not None:
            log.info("Saving generated flann index to: %s",
                     flann_index_file)
            flann.save_index(flann_index_file)
        else:
            log.warning("Constructed new index file "
                        "but did not save it.")

    else:
        log.debug("Given existing index file. Loading it.")
        flann.load_index(flann_index_file, codebook)

    return flann


# noinspection PyUnusedLocal
def flann_quantize_data(flann,
                        filein_name, fileout_name,
                        func_normalize,
                        filein_gzipped=False, filein_bzipped=False,
                        delimiter=' ',
                        fileout_gzipped=False, fileout_bzipped=False,
                        k=10, sparsity=1, size_block=20000):
    """
    Quantize raw data based on flann module

    @param flann: FLANN index instance.
    @param filein_name: input file (raw contents before quantization)
    @param fileout_name: output file (quantized results)
    @param func_normalize: a function to normalize each data row (if not
        necessary, use None)
    @param filein_gzipped: is the input file in gziped encoding?
        (default = False)
    @param filein_bzipped:is the input file in bziped encoding?
        (default = False)
    @param delimiter: delimiter used for input file
    @param fileout_gzipped: if the file out should be compressed with gzip
    @param fileout_bzipped: if the file out should be compressed with bzip
    @param k: number of quantization nearest neighbors to be saved (if k>1,
        that's soft quantization)
    @param sparsity: controls the amount of input data to be used. E.g., if 3,
        only every three lines are used.
    @param size_block: number of data points to be used at once. Need to be
        adjusted baesd on memory and CPU IO speed
    """
    assert not (filein_gzipped and filein_bzipped), \
        "Input file cannot be declared as both gzipped and bzipped."
    assert not (fileout_gzipped and fileout_bzipped), \
        "Output file cannot be declared as both gzipped and bzipped."

    if filein_gzipped or os.path.splitext(filein_name)[1] == '.gz':
        fin = gzip.open(filein_name, 'rb')
    elif filein_bzipped or os.path.splitext(filein_name)[1] == '.bz2':
        fin = bz2.BZ2File(filein_name, 'rb')
    else:
        fin = open(filein_name, 'rb')

    # tracks how many lines are read from file
    count1 = 0
    # tracks how many lines are to be processed by being stored in 'lines'
    count2 = 0
    lines = []

    fout = None
    if fileout_gzipped or os.path.splitext(fileout_name)[1] == '.gz':
        fout = gzip.open(fileout_name, 'w')
    elif fileout_bzipped or os.path.splitext(fileout_name)[1] == '.bz2':
        fout = bz2.BZ2File(fileout_name, 'w')
    else:
        fout = open(fileout_name, 'w')

    def quantize_then_write(lines_in):
        data = np.array(lines_in)
        data = np.float64(data)

        if func_normalize:
            data[:, 3:] = func_normalize(data[:, 3:])

        idx, dists = flann.nn_index(data[:, 3:], k)
        out_data = np.concatenate((data[:, 0:3], idx, dists), axis=1)
        np.savetxt(fout, out_data, fmt='%g')

    for line in csv.reader(fin, delimiter=' '):
        count1 += 1
        if sparsity != 1 and (count1 % sparsity) == 0:
            continue

        lines.append(line)
        count2 += 1

        if count2 == size_block:
            quantize_then_write(lines)
            count2 = 0  # reset
            lines = []

    if count2 > 0:
        quantize_then_write(lines)

    fin.close()
    fout.close()

    return 1


def flann_quantize_data2(flann, in_descriptors, func_normalize=None,
                         k=10, sparsity=1, size_block=20000):
    """
    Quantize raw data based on flan module (in-memory version)

    :param flann: Prepared FLANN module
    :type flann: pyflann.FLANN
    :param in_descriptors: descriptors matrix (colorDescriptor)
    :type in_descriptors: numpy.matrixlib.defmatrix.matrix
    :param func_normalize: a function to normalize each data row (if not
        necessary, use None)
    :param k: number of quantization nearest neighbors to be saved (if k>1,
        that's soft quantization)
    :param sparsity: controls the amount of input data to be used. E.g., if 3,
        only every three lines are used.
    :param size_block: number of data points to be used at once. Need to be
        adjusted baesd on memory and CPU IO speed

    :return:
    :rtype:

    """
    # Input descriptor matrix is of the format:
    # [[ frm, x, y, descriptor... ]
    #  [ ...
    #  ...

    # tracks how many lines are read from file
    count1 = 0

    # Get array base, better suited for loop below
    in_descriptors = in_descriptors.A

    rows = []
    quantized = None

    # Quantization helper function
    # Finds k nearest neighbors to descriptor, creating a vector that is of the
    # format:
    #   [ frm x y nn_1 ... nn_k dist_1 ... dist_k ]
    # which has a total size of (k*2)+3
    # noinspection PyShadowingNames
    def quantize_then_write(rows, quantized):
        data = np.matrix(rows)

        if func_normalize:
            data[:, 3:] = func_normalize(data[:, 3:])

        idx, dists = flann.nn_index(data[:, 3:], k)
        out_data = np.concatenate((data[:, 0:3], idx, dists), axis=1)

        if quantized is None:
            quantized = out_data
        else:
            np.vstack((quantized, out_data))
        return quantized

    # Run through rows in input matrix
    for row in in_descriptors:
        count1 += 1
        if sparsity > 1 and (count1 % sparsity) == 0:
            continue

        rows.append(row)

        if len(rows) == size_block:
            quantized = quantize_then_write(rows, quantized)
            rows = []

    if len(rows) > 0:
        quantized = quantize_then_write(rows, quantized)

    # Output Format:
    # [[ frame, x, y, idx0, ..., idx9, dist0, ..., dist9 ]
    #  [ ... [
    #  ...
    return quantized


def quantizeResults(key, outdir, outtype,
                    pattern_codebook='%s_codebook_med12.txt'):
    # feature file for quantization
    filein_name = os.path.join(outdir, key+'.'+outtype+'-all.txt')
    outname = outtype + '-all.encode.txt'
    fileout_name = os.path.join(outdir, key+'.'+outname)
    # put codebook file under the path
    thisdir = os.path.dirname(os.path.realpath(__file__))
    file_codebook = os.path.join(thisdir, pattern_codebook % outtype)
    # put flann index file under the path
    file_flann = os.path.join(thisdir, '%s.flann' % outtype)
    # load codebook and build flann knn query engine
    cbook = flann_load_codebook(file_codebook, is_rowwise=False)
    flann = flann_build(cbook, file_flann)
    # quantize the data
    flann_quantize_data(flann, filein_name, fileout_name, None)
    flann.delete_index()


def quantizeResults2(file_input, file_output, file_codebook, file_flann,
                     filein_gzipped=False, filein_bzipped=False):
    """
    Another version of raw feature quantization using FLANN.
    No directory structure is assumed. Simply input/ouput/parameter files are
    given.

    @param file_input: Input raw file with first three columns representing
        feature location info
    @param file_output: quantization file to be generated
    @param file_codebook: codebook file
    @param file_flann: quantization parameters for approximate nearest-neighbor
    @param filein_gzipped: if True, treat the file as gzipped version
    @param filein_bzipped: if True, treat the file as bzipped version
    """
    cbook = flann_load_codebook(file_codebook, is_rowwise=False)
    flann = flann_build(cbook, file_flann)

    flann_quantize_data(flann, file_input, file_output, None,
                        filein_gzipped=filein_gzipped,
                        filein_bzipped=filein_bzipped)
    flann.delete_index()


def quantizeResults3(in_descriptors, file_codebook, file_flann):
    """
    Another version of raw feature quantization using FLANN.
    No directory structure is assumed. Uses input numpy matrix vs. file.

    :param in_descriptors: Matrix of input descriptors, first row: frame number,
        rows 2-3: colorDescriptor info, rows 4+: descriptor
    :param file_codebook: codebook file
    :param file_flann: quantization parameters for approximate nearest-neighbor

    """
    cbook = flann_load_codebook(file_codebook, is_rowwise=False)
    flann = flann_build(cbook, file_flann)

    quantized = flann_quantize_data2(flann, in_descriptors)
    flann.delete_index()
    return quantized


#########################################################################
def build_sp_hist(key, outdir, outtype):
    inname = outtype + '-all.encode.txt'
    filein = os.path.join(outdir, key+'.'+inname)
    outname = outtype + '-sp.hist.txt'
    fileout = os.path.join(outdir, key+'.'+outname)
    return build_sp_hist_(filein, fileout)


def build_sp_hist_(filein, fileout, bins_code=np.arange(0, 4096 + 1)):
    feas = np.loadtxt(filein)
    cordx = feas[:, 1]
    cordy = feas[:, 2]
    feas = feas[:, 3]  # only the top component, we are looking
    # hard quantization
    # global histogram
    #: :type: numpy.core.multiarray.ndarray
    hist_csift_g = np.histogram(feas, bins=bins_code)[0]
    hist_csift_g = hist_csift_g[np.newaxis]
    # 4 quadrants
    midx = np.ceil(cordx.max()/2)
    midy = np.ceil(cordy.max()/2)
    lx = cordx < midx
    rx = cordx >= midx
    uy = cordy < midy
    dy = cordy >= midy
    # logging.debug("LXUI: %s,%s" % (lx.__repr__(), uy.__repr__()))
    # logging.debug("Length LXUI: %s,%s " % (lx.shape, uy.shape))
    # logging.debug("feas dimensions: %" % (feas.shape,))

    #: :type: numpy.core.multiarray.ndarray
    hist_csift_q1 = np.histogram(feas[lx & uy], bins=bins_code)[0]
    #: :type: numpy.core.multiarray.ndarray
    hist_csift_q2 = np.histogram(feas[rx & uy], bins=bins_code)[0]
    #: :type: numpy.core.multiarray.ndarray
    hist_csift_q3 = np.histogram(feas[lx & dy], bins=bins_code)[0]
    #: :type: numpy.core.multiarray.ndarray
    hist_csift_q4 = np.histogram(feas[rx & dy], bins=bins_code)[0]
    hist_csift_q1 = hist_csift_q1[np.newaxis]
    hist_csift_q2 = hist_csift_q2[np.newaxis]
    hist_csift_q3 = hist_csift_q3[np.newaxis]
    hist_csift_q4 = hist_csift_q4[np.newaxis]

    # 3 layers
    ythird = np.ceil(cordy.max()/3)
    l1 = cordy <= ythird
    l2 = (cordy > ythird) & (cordy <= 2*ythird)
    l3 = cordy > 2*ythird
    #: :type: numpy.core.multiarray.ndarray
    hist_csift_l1 = np.histogram(feas[l1], bins=bins_code)[0]
    #: :type: numpy.core.multiarray.ndarray
    hist_csift_l2 = np.histogram(feas[l2], bins=bins_code)[0]
    #: :type: numpy.core.multiarray.ndarray
    hist_csift_l3 = np.histogram(feas[l3], bins=bins_code)[0]
    hist_csift_l1 = hist_csift_l1[np.newaxis]
    hist_csift_l2 = hist_csift_l2[np.newaxis]
    hist_csift_l3 = hist_csift_l3[np.newaxis]
    # concatenate
    hist_csift = np.vstack((hist_csift_g, hist_csift_q1, hist_csift_q2,
                            hist_csift_q3, hist_csift_q4, hist_csift_l1,
                            hist_csift_l2, hist_csift_l3))
    np.savetxt(fileout, hist_csift, fmt='%g')
    return hist_csift


def build_sp_hist2(feas, bins_code=np.arange(0, 4096+1)):
    """ Build spacial pyramid from quantized data

    :param feas: quantized data matrix
    :type feas: numpy.matrixlib.defmatrix.matrix

    :param bins_code: Index code array for input features.  This is usually a
        sequential array of integers the same size as the dimensionalty of the
        feature vectors.
    :type bins_code: np.ndarray[int]

    :return: martrix of 8 rows representing the histograms for the different
        spacial regions.
    :rtype: numpy.matrixlib.defmatrix.matrix

    """
    cordx = feas[:, 1]
    cordy = feas[:, 2]
    feas = feas[:, 3]  # only the top component, we are looking
    # hard quantization
    # global histogram
    #: :type: numpy.core.multiarray.ndarray
    hist_csift_g = np.histogram(feas, bins=bins_code)[0]
    hist_csift_g = hist_csift_g[np.newaxis]
    # 4 quadrants
    # noinspection PyTypeChecker
    midx = np.ceil(cordx.max()/2)
    # noinspection PyTypeChecker
    midy = np.ceil(cordy.max()/2)
    lx = cordx < midx
    rx = cordx >= midx
    uy = cordy < midy
    dy = cordy >= midy
    # logging.error("LXUI: %s,%s", lx.__repr__(), uy.__repr__())
    # logging.error("Length LXUI: %s,%s", lx.shape, uy.shape)
    # logging.error("feas dimensions: %s", feas.shape)

    #: :type: numpy.core.multiarray.ndarray
    hist_csift_q1 = np.histogram(feas[lx & uy], bins=bins_code)[0]
    #: :type: numpy.core.multiarray.ndarray
    hist_csift_q2 = np.histogram(feas[rx & uy], bins=bins_code)[0]
    #: :type: numpy.core.multiarray.ndarray
    hist_csift_q3 = np.histogram(feas[lx & dy], bins=bins_code)[0]
    #: :type: numpy.core.multiarray.ndarray
    hist_csift_q4 = np.histogram(feas[rx & dy], bins=bins_code)[0]
    hist_csift_q1 = hist_csift_q1[np.newaxis]
    hist_csift_q2 = hist_csift_q2[np.newaxis]
    hist_csift_q3 = hist_csift_q3[np.newaxis]
    hist_csift_q4 = hist_csift_q4[np.newaxis]

    # 3 layers
    # noinspection PyTypeChecker
    ythird = np.ceil(cordy.max()/3)
    l1 = cordy <= ythird
    l2 = (cordy > ythird) & (cordy <= 2*ythird)
    l3 = cordy > 2*ythird
    #: :type: numpy.core.multiarray.ndarray
    hist_csift_l1 = np.histogram(feas[l1], bins=bins_code)[0]
    #: :type: numpy.core.multiarray.ndarray
    hist_csift_l2 = np.histogram(feas[l2], bins=bins_code)[0]
    #: :type: numpy.core.multiarray.ndarray
    hist_csift_l3 = np.histogram(feas[l3], bins=bins_code)[0]
    hist_csift_l1 = hist_csift_l1[np.newaxis]
    hist_csift_l2 = hist_csift_l2[np.newaxis]
    hist_csift_l3 = hist_csift_l3[np.newaxis]
    # concatenate
    hist_csift = np.vstack((hist_csift_g, hist_csift_q1, hist_csift_q2,
                            hist_csift_q3, hist_csift_q4, hist_csift_l1,
                            hist_csift_l2, hist_csift_l3))
    return hist_csift



"""
Add a set of local system files to a data set via explicit paths or shell-style
glob strings.
"""

import glob
import logging
import os.path as osp

from smqtk.representation import get_data_set_impls
from smqtk.representation.data_element.file_element import DataFileElement
from smqtk.utils import bin_utils, plugin


def default_config():
    return {
        "data_set": plugin.make_config(get_data_set_impls())
    }


def cli_parser():
    parser = bin_utils.basic_cli_parser(__doc__)
    parser.add_argument("input_files", metavar='GLOB', nargs='*')
    return parser


def main():
    parser = cli_parser()
    args = parser.parse_args()
    config = bin_utils.utility_main_helper(default_config, args)
    log = logging.getLogger(__name__)

    log.debug("Script arguments:\n%s" % args)

    def iter_input_elements():
        for f in args.input_files:
            f = osp.expanduser(f)
            if osp.isfile(f):
                yield DataFileElement(f)
            else:
                log.debug("Expanding glob: %s" % f)
                for g in glob.glob(f):
                    yield DataFileElement(g)

    log.info("Adding elements to data set")
    #: :type: smqtk.representation.DataSet
    ds = plugin.from_plugin_config(config['data_set'], get_data_set_impls())
    ds.add_data(*iter_input_elements())


if __name__ == '__main__':
    main()



import heapq
import threading

from six import BytesIO
from six.moves import map

import numpy

from smqtk.algorithms.nn_index.hash_index import HashIndex
from smqtk.representation import get_data_element_impls
from smqtk.utils import merge_dict, plugin
from smqtk.utils.bit_utils import (
    bit_vector_to_int_large,
    int_to_bit_vector_large,
)
from smqtk.utils.metrics import hamming_distance


class LinearHashIndex (HashIndex):
    """
    Basic linear index using heap sort (aka brute force).
    Hash codes are stored as large integer values.
    """

    @classmethod
    def is_usable(cls):
        return True

    @classmethod
    def get_default_config(cls):
        """
        Generate and return a default configuration dictionary for this class.
        This will be primarily used for generating what the configuration
        dictionary would look like for this class without instantiating it.

        By default, we observe what this class's constructor takes as arguments,
        turning those argument names into configuration dictionary keys. If any
        of those arguments have defaults, we will add those values into the
        configuration dictionary appropriately. The dictionary returned should
        only contain JSON compliant value types.

        It is not be guaranteed that the configuration dictionary returned
        from this method is valid for construction of an instance of this class.

        :return: Default configuration dictionary for the class.
        :rtype: dict

        """
        c = super(LinearHashIndex, cls).get_default_config()
        c['cache_element'] = plugin.make_config(get_data_element_impls())
        return c

    @classmethod
    def from_config(cls, config_dict, merge_default=True):
        """
        Instantiate a new instance of this class given the configuration
        JSON-compliant dictionary encapsulating initialization arguments.

        This method should not be called via super unless an instance of the
        class is desired.

        :param config_dict: JSON compliant dictionary encapsulating
            a configuration.
        :type config_dict: dict

        :param merge_default: Merge the given configuration on top of the
            default provided by ``get_default_config``.
        :type merge_default: bool

        :return: Constructed instance from the provided config.
        :rtype: LinearHashIndex

        """
        if merge_default:
            config_dict = merge_dict(cls.get_default_config(), config_dict)

        cache_element = None
        if config_dict['cache_element'] \
                and config_dict['cache_element']['type']:
            cache_element = \
                plugin.from_plugin_config(config_dict['cache_element'],
                                          get_data_element_impls())
        config_dict['cache_element'] = cache_element

        return super(LinearHashIndex, cls).from_config(config_dict, False)

    def __init__(self, cache_element=None):
        """
        Initialize linear, brute-force hash index.

        :param cache_element: Optional data element to cache our index to.
        :type cache_element: smqtk.representation.DataElement | None

        """
        super(LinearHashIndex, self).__init__()
        self.cache_element = cache_element
        # Our index is the set of bit-vectors as an integers/longs.
        #: :type: set[int]
        self.index = set()
        self._model_lock = threading.RLock()
        self.load_cache()

    def get_config(self):
        c = self.get_default_config()
        if self.cache_element:
            c['cache_element'] = merge_dict(c['cache_element'],
                                            plugin.to_plugin_config(
                                                self.cache_element))
        return c

    def load_cache(self):
        """
        Load from file cache if we have one
        """
        with self._model_lock:
            if self.cache_element and not self.cache_element.is_empty():
                buff = BytesIO(self.cache_element.get_bytes())
                self.index = set(numpy.load(buff))

    def save_cache(self):
        """
        save to file cache if configures
        """
        with self._model_lock:
            if self.cache_element and self.index:
                if self.cache_element.is_read_only():
                    raise ValueError("Cache element (%s) is read-only."
                                     % self.cache_element)
                buff = BytesIO()
                # noinspection PyTypeChecker
                numpy.save(buff, tuple(self.index))
                self.cache_element.set_bytes(buff.getvalue())

    def count(self):
        with self._model_lock:
            return len(self.index)

    def _build_index(self, hashes):
        """
        Internal method to be implemented by sub-classes to build the index with
        the given hash codes (bit-vectors).

        Subsequent calls to this method should rebuild the current index.  This
        method shall not add to the existing index nor raise an exception to as
        to protect the current index.

        :param hashes: Iterable of descriptor elements to build index
            over.
        :type hashes: collections.Iterable[numpy.ndarray[bool]]

        """
        with self._model_lock:
            new_index = set(map(bit_vector_to_int_large, hashes))
            self.index = new_index
            self.save_cache()

    def _update_index(self, hashes):
        """
        Internal method to be implemented by sub-classes to additively update
        the current index with the one or more hash vectors given.

        If no index exists yet, a new one should be created using the given hash
        vectors.

        :param hashes: Iterable of numpy boolean hash vectors to add to this
            index.
        :type hashes: collections.Iterable[numpy.ndarray[bool]]

        """
        with self._model_lock:
            self.index.update(set(map(bit_vector_to_int_large, hashes)))
            self.save_cache()

    def _remove_from_index(self, hashes):
        """
        Internal method to be implemented by sub-classes to partially remove
        hashes from this index.

        :param hashes: Iterable of numpy boolean hash vectors to remove from
            this index.
        :type hashes: collections.Iterable[numpy.ndarray[bool]]

        :raises KeyError: One or more hashes provided do not match any stored
            hashes.  The index should not be modified.

        """
        with self._model_lock:
            h_int_set = set(map(bit_vector_to_int_large, hashes))
            # KeyError if any hash ints are not in our index map.
            for h in h_int_set:
                if h not in self.index:
                    raise KeyError(h)
            self.index = self.index - h_int_set
            self.save_cache()

    def _nn(self, h, n=1):
        """
        Internal method to be implemented by sub-classes to return the nearest
        `N` neighbor hash codes as bit-vectors to the given hash code
        bit-vector.

        Distances are in the range [0,1] and are the percent different each
        neighbor hash is from the query, based on the number of bits contained
        in the query (normalized hamming distance).

        When this internal method is called, we have already checked that our
        index is not empty.

        :param h: Hash code to compute the neighbors of. Should be the same bit
            length as indexed hash codes.
        :type h: numpy.ndarray[bool]

        :param n: Number of nearest neighbors to find.
        :type n: int

        :return: Tuple of nearest N hash codes and a tuple of the distance
            values to those neighbors.
        :rtype: (tuple[numpy.ndarray[bool]], tuple[float])

        """
        with self._model_lock:
            h_int = bit_vector_to_int_large(h)
            bits = len(h)
            #: :type: list[int|long]
            near_codes = \
                heapq.nsmallest(n, self.index,
                                lambda e: hamming_distance(h_int, e)
                                )
            distances = map(hamming_distance, near_codes,
                            [h_int] * len(near_codes))
            return [int_to_bit_vector_large(c, bits) for c in near_codes], \
                   [d / float(bits) for d in distances]



import unittest

from smqtk.utils.configurable_interface import Configurable


class TestConfigurablePluginInterface (unittest.TestCase):

    def test_default_config(self):
        # Only using class method.
        # noinspection PyAbstractClass
        class T (Configurable):
            # noinspection PyUnusedLocal
            def __init__(self, a, b, cat):
                pass
        self.assertEqual(
            T.get_default_config(),
            {'a': None, 'b': None, 'cat': None}
        )

    def test_default_config_with_star_args(self):
        # Test that star stuff shouldn't change anything

        # Only using class method.
        # noinspection PyAbstractClass
        class T (Configurable):
            # noinspection PyUnusedLocal
            def __init__(self, a, b, cat, *args, **kwargs):
                pass
        self.assertEqual(
            T.get_default_config(),
            {'a': None, 'b': None, 'cat': None}
        )

    def test_default_config_with_default_values(self):
        # Only using class method.
        # noinspection PyAbstractClass
        class T (Configurable):
            # noinspection PyUnusedLocal
            def __init__(self, a, b, c=0, d='foobar'):
                pass
        self.assertEqual(
            T.get_default_config(),
            {'a': None, 'b': None, 'c': 0, 'd': 'foobar'}
        )

    def test_default_config_with_default_values_with_star_args(self):
        # Test that star stuff shouldn't change anything.

        # Only using class method.
        # noinspection PyAbstractClass
        class T (Configurable):
            # noinspection PyUnusedLocal
            def __init__(self, a, b, c=0, d='foobar', *args, **kwargs):
                pass
        self.assertEqual(
            T.get_default_config(),
            {'a': None, 'b': None, 'c': 0, 'd': 'foobar'}
        )


class DummyAlgo1 (Configurable):

    def __init__(self, foo=1, bar='baz'):
        self.foo = foo
        self.bar = bar

    def __eq__(self, other):
        return self.foo == other.foo and self.bar == other.bar

    def __ne__(self, other):
        return not (self == other)

    def get_config(self):
        return {
            'foo': self.foo,
            'bar': self.bar,
        }


class DummyAlgo2 (Configurable):
    """
    Semi-standard way to implement nested algorithms.
    Usually, algorithms are associated to a plugin getter method that wraps a
    call to ``smqtk.utils.plugin.get_plugins``, but I'm skipping that for now
    for simplicity.
    """

    @classmethod
    def get_default_config(cls):
        default = super(DummyAlgo2, cls).get_default_config()
        # replace ``child`` with config available config
        default['child'] = default['child'].get_config()
        return default

    @classmethod
    def from_config(cls, config, merge_default=True):
        config['child'] = DummyAlgo1(**config['child'])
        return cls(**config)

    def __init__(self, child=DummyAlgo1(), alpha=0.0001, beta='default'):
        # Where child is supposed to be an instance of DummyAlgo1
        self.child = child
        self.alpha = alpha
        self.beta = beta

    def __eq__(self, other):
        return (
            self.child == other.child and
            self.alpha == other.alpha and
            self.beta == other.beta
        )

    def get_config(self):
        return {
            'child': self.child.get_config(),
            'alpha': self.alpha,
            'beta': self.beta
        }


class TestConfigurablePluginTestImpl (unittest.TestCase):

    def test_algo1_default(self):
        self.assertEqual(
            DummyAlgo1.get_default_config(),
            {'foo': 1, 'bar': 'baz'}
        )

    def test_algo2_default(self):
        self.assertEqual(
            DummyAlgo2.get_default_config(),
            {
                'child': {'foo': 1, 'bar': 'baz'},
                'alpha': 0.0001,
                'beta': 'default'
            }
        )



"""
LICENCE
-------
Copyright 2013 by Kitware, Inc. All Rights Reserved. Please refer to
KITWARE_LICENSE.TXT for licensing information, or contact General Counsel,
Kitware, Inc., 28 Corporate Drive, Clifton Park, NY 12065.


Object for handling signal overrides and managing what signals have been seen
since override.

"""

import logging
import re
import signal
import threading


class SignalHandler(object):
    """
    Object for handling knowledge of whether a particular system signal has
    been caught for the current thread of execution.

    This object is thread safe and acts in a synchronous manner.
    """

    def __init__(self, name='SignalHandler'):
        self.__signal_caught = {}
        self.__s_lock = threading.Lock()
        self.__prev_handlers = {}
        self.__h_lock = threading.Lock()

        self._log = logging.getLogger(__name__ + '.' + name)

        # create a mapping of signal integer keys to the signal they represent
        # this may change from system to system(?), thus the creating it
        # automatically
        sig_list = [
            re.match('^SIG[A-Z]+$', e) and re.match('^SIG[A-Z]+$', e).group()
            for e in dir(signal)
        ]
        sig_list.sort()
        sig_list = sig_list[sig_list.count(None):]
        self._sig_map = dict([(getattr(signal, sig), sig)
                              for sig in sig_list])

    def _gen_signal_handle(self, custom_func=None):
        """
        Generated a callback method to be registered to signal.signal for a
        particular signal to catch.

        :param custom_func: A custom signal handle function to add
            functionality to our handler.
        :type custom_func: (int, None|frame) -> None

        :return: Function handle method for registering.
        :rtype: types.FunctionType

        """
        def handle_signal(signum, stack):
            """
            Callback method to be registered to signal.signal for a particular
            signal to catch.

            :type signum: int
            :type stack: None | frame

            """
            with self.__s_lock:
                self._log.debug("'%s' caught", self._sig_map[signum])
                self.__signal_caught[signum] = True

            if custom_func:
                custom_func(signum, stack)

        return handle_signal

    def register_signal(self, signum, custom_func=None):
        """
        Register a signal to be handled and monitored if not already
        registered.
        We will override and record the previous handler so that we may put it
        back if/when unregister that signal.

        A custom handling function may be passed to extend actions taken when
        the signal is caught. This method should be of the form of a normal
        signal handle method (see python docs). Regardless if a custom function
        is given, we will always register in our map that a handler has been
        registered for that signal, preventing another handle to be registered
        for that signal until the current one is removed, as well as still
        registering that the signal has been caught.

        If we have already registered the given signal, we return False.

        :type signum: int
        :param signum: The identifying integer value of the signal to check.
        (use signal.SIGINT, signal.SIGTERM, etc.)

        :param custom_func: Custom callback function to be called when the
            specified signal is caught. This function must take two arguments
            where the first is the integer signal identifier and the second
            is the stack frame in which the signal was caught.
        :type custom_func: (int, frame) -> None

        :return: True of we successfully registered a new signal, or False if
        the signal is already registered.

        """
        # if we haven't already registered a handler for the given signal.
        with self.__h_lock:
            if signum not in self.__prev_handlers:
                self._log.debug("Registering catch for signal %i (%s)",
                                signum, self._sig_map[signum])
                prev_handle = \
                    signal.signal(signum, self._gen_signal_handle(custom_func))
                self.__prev_handlers[signum] = prev_handle
                return True
            self._log.debug("%s already registered", self._sig_map[signum])
            return False

    def unregister_signal(self, signum):
        """
        Unregister the given signal to the previous handler we have recorded.
        This also sets whether the signal has been caught or not to False.

        If the given signal is not registered, we return False.

        :type signum: int
        :param signum: The identifying integer value of the signal to check.
        (use signal.SIGINT, signal.SIGTERM, etc.)

        return: True if we successfully unregistered a signal, or False if the
        the given signal was not registered in the first place.

        """
        with self.__h_lock:
            if signum in self.__prev_handlers:
                self._log.debug("Restoring previous handler for %s",
                                self._sig_map[signum])
                signal.signal(signum, self.__prev_handlers[signum])
                del self.__prev_handlers[signum]
                # unregister signal caught value
                self.reset_signal(signum)
                return True
            self._log.debug("%s never registered.", self._sig_map[signum])
            return False

    def signals_registered(self):
        """
        Return what signals are currently registered in this handler

        :return: what signals are currently registered in this handler
        :rtype: list of int

        """
        return list(self.__prev_handlers.keys())

    def reset_signal(self, signum):
        """
        Reset our knowledge of whether a particular signal has been caught.
        This does NOT unregister any handlers.

        If the given signal is not registered, we return False.

        :type signum: int
        :param signum: The identifying integer value of the signal to check.
        (use signal.SIGINT, signal.SIGTERM, etc.)

        :return: True if the signal catch record was reset, and False if the
        given signal is not registered as being monitored.

        """
        with self.__s_lock:
            if signum in self.__signal_caught:
                self._log.debug("Resetting %s catch boolean",
                                self._sig_map[signum])
                del self.__signal_caught[signum]
                return True
            self._log.debug("%s never caught, nothing to reset",
                            self._sig_map[signum])
            return False

    def reset(self):
        """
        Reset our knowledge of what signals have been caught. This does NOT
        unregister any handlers.

        """
        with self.__s_lock:
            self._log.debug("Resetting all signal catches")
            self.__signal_caught = {}

    def is_signal_caught(self, signum):
        """
        Check if we have caught the given signal since the creation of the
        object or since the last reset. We also return false if the given
        signal isn't being monitored.

        :type signum: int
        :param signum: The identifying integer value of the signal to check.
        (use signal.SIGINT, signal.SIGTERM, etc.)

        :rtype: bool
        :return; True if the signal has been seen since creation/reset, or
            False otherwise.

        """
        return self.__signal_caught.get(signum, False)



"""
Script for generating clusters from descriptors in a given index using the
mini-batch KMeans implementation from Scikit-learn
(http://scikit-learn.org/stable/modules/generated/sklearn.cluster.MiniBatchKMeans.html).

By the nature of Scikit-learn's MiniBatchKMeans implementation, euclidean
distance is used to measure distance between descriptors.
"""
import logging
import os

import numpy
from six.moves import cPickle
from sklearn.cluster import MiniBatchKMeans

from smqtk.compute_functions import mb_kmeans_build_apply
from smqtk.representation.descriptor_index import get_descriptor_index_impls
from smqtk.utils import Configurable
from smqtk.utils.bin_utils import utility_main_helper, basic_cli_parser
from smqtk.utils.file_utils import safe_create_dir
from smqtk.utils.plugin import make_config, from_plugin_config


def default_config():

    # Trick for mixing in our Configurable class API on top of scikit-learn's
    # MiniBatchKMeans class in order to introspect construction parameters.
    # We never construct this class so we do not need to implement "pure
    # virtual" instance methods.
    # noinspection PyAbstractClass
    class MBKTemp (MiniBatchKMeans, Configurable):
        pass

    c = {
        "minibatch_kmeans_params": MBKTemp.get_default_config(),
        "descriptor_index": make_config(get_descriptor_index_impls()),
        # Number of descriptors to run an initial fit with. This brings the
        # advantage of choosing a best initialization point from multiple.
        "initial_fit_size": 0,
        # Path to save generated KMeans centroids
        "centroids_output_filepath_npy": "centroids.npy"
    }

    # Change/Remove some KMeans params for more appropriate defaults
    del c['minibatch_kmeans_params']['compute_labels']
    del c['minibatch_kmeans_params']['verbose']
    c['minibatch_kmeans_params']['random_state'] = 0

    return c


def cli_parser():
    p = basic_cli_parser(__doc__)

    g_output = p.add_argument_group("output")
    g_output.add_argument('-o', '--output-map',
                          metavar="PATH",
                          help="Path to output the clustering class mapping "
                               "to. Saved as a pickle file with -1 format.")

    return p


def main():
    args = cli_parser().parse_args()
    config = utility_main_helper(default_config, args)
    log = logging.getLogger(__name__)

    output_filepath = args.output_map
    if not output_filepath:
        raise ValueError("No path given for output map file (pickle).")

    #: :type: smqtk.representation.DescriptorIndex
    index = from_plugin_config(config['descriptor_index'],
                               get_descriptor_index_impls())
    mbkm = MiniBatchKMeans(verbose=args.verbose,
                           compute_labels=False,
                           **config['minibatch_kmeans_params'])
    initial_fit_size = int(config['initial_fit_size'])

    d_classes = mb_kmeans_build_apply(index, mbkm, initial_fit_size)

    log.info("Saving KMeans centroids to: %s",
             config['centroids_output_filepath_npy'])
    numpy.save(config['centroids_output_filepath_npy'], mbkm.cluster_centers_)

    log.info("Saving result classification map to: %s", output_filepath)
    safe_create_dir(os.path.dirname(output_filepath))
    with open(output_filepath, 'w') as f:
        cPickle.dump(d_classes, f, -1)

    log.info("Done")


if __name__ == '__main__':
    main()



from __future__ import division, print_function
import unittest

from smqtk.utils.string_utils import random_characters


class TestRandomCharacters (unittest.TestCase):

    def test_zero_n(self):
        self.assertEqual(random_characters(0), '')

    def test_no_char_set(self):
        self.assertRaisesRegexp(
            ValueError,
            "Empty char_set given",
            random_characters, 5, ()
        )

    def test_negative_n(self):
        self.assertRaisesRegexp(
            ValueError,
            "n must be a positive integer",
            random_characters, -1234
        )

    def test_floating_point_n(self):
        # Testing that this should cast down to the integer 4.
        # noinspection PyTypeChecker
        s = random_characters(4.7)
        self.assertEqual(len(s), 4)



"""
LICENCE
-------
Copyright 2013 by Kitware, Inc. All Rights Reserved. Please refer to
KITWARE_LICENSE.TXT for licensing information, or contact General Counsel,
Kitware, Inc., 28 Corporate Drive, Clifton Park, NY 12065.

"""


class DatabaseInfo (object):
    """
    Encapsulation of database connection information meant for MongoDB. This
    may or may not have collection information (None if it doesn't).
    """

    def __init__(self, host, port, name):
        """
        Create a database info encapsulation object

        >>> dbi = DatabaseInfo('localhost', 12345, 'foobar')
        >>> dbi.host
        'localhost'
        >>> dbi.port
        12345
        >>> dbi.name
        'foobar'

        :param host:
        :type host:
        :param port:
        :type port:
        :param name:
        :type name:
        :return:
        :rtype:
        """
        self.host = str(host)
        self.port = int(port)
        self.name = str(name)

    def copy(self):
        """
        Create an exact duplicate of this object. This is a deep copy.

        >>> dbi = DatabaseInfo('localhost', 12345, 'foobar')
        >>> dbi2 = dbi.copy()
        >>> dbi.host == dbi2.host
        True
        >>> dbi.port == dbi2.port
        True
        >>> dbi.name == dbi2.name
        True

        :return: a completely new DatabaseInfo instance with the same data as
            this one.
        :rtype: DatabaseInfo

        """
        return DatabaseInfo(self.host, self.port, self.name)

    def __repr__(self):
        """
        >>> dbi = DatabaseInfo('localhost', 12345, 'foobar')
        >>> print(dbi)
        DatabaseInfo{host: localhost, port: 12345, name: foobar}

        """
        return "DatabaseInfo{host: %s, port: %d, name: %s}" \
               % (self.host, self.port, self.name)



from __future__ import division, print_function
import random
import unittest

import numpy
# noinspection PyUnresolvedReferences
from six.moves import range

from smqtk.utils import bit_utils


class TestBitUtils (unittest.TestCase):

    def test_int_to_bit_vector_large_0(self):
        # Need at least one bit to represent 0.
        numpy.testing.assert_array_equal(
            bit_utils.int_to_bit_vector_large(0),
            [False]
        )
        # Force 5 bits.
        numpy.testing.assert_array_equal(
            bit_utils.int_to_bit_vector_large(0, 5),
            [False, False, False, False, False]
        )

    def test_int_to_bit_vector_large_1(self):
        numpy.testing.assert_array_equal(
            bit_utils.int_to_bit_vector_large(1),
            [True]
        )
        numpy.testing.assert_array_equal(
            bit_utils.int_to_bit_vector_large(1, 7),
            ([False] * 6) + [True]
        )

    def test_int_to_bit_vector_large_large(self):
        # Try large integer bit vectors
        int_val = (2**256) - 1
        expected_vector = [True] * 256
        numpy.testing.assert_array_equal(
            bit_utils.int_to_bit_vector_large(int_val),
            expected_vector
        )

        int_val = (2**512)
        expected_vector = [True] + ([False] * 512)
        numpy.testing.assert_array_equal(
            bit_utils.int_to_bit_vector_large(int_val),
            expected_vector
        )

    def test_int_to_bit_vector_large_invalid_bits(self):
        # Cannot represent 5 in binary using 1 bit.
        self.assertRaises(
            ValueError,
            bit_utils.int_to_bit_vector_large,
            5, 1
        )

    def test_popcount(self):
        self.assertEqual(bit_utils.popcount(1), 1)
        self.assertEqual(bit_utils.popcount(2), 1)
        self.assertEqual(bit_utils.popcount(3), 2)
        self.assertEqual(bit_utils.popcount(2 ** 16), 1)
        self.assertEqual(bit_utils.popcount(2**16 - 1), 16)
        self.assertEqual(bit_utils.popcount(2 ** 32), 1)
        self.assertEqual(bit_utils.popcount(2 ** 32 - 1), 32)

    def test_popcount_0(self):
        self.assertEqual(bit_utils.popcount(0), 0)

    def test_popcount_limits(self):
        # Make sure documented integer limit is truthful.
        c = 10000
        for _ in range(c):
            # noinspection PyUnresolvedReferences
            v = random.randint(0, bit_utils.popcount.v_max)
            # Known method to always work based on counting python's binary
            # string representation.
            v_bin_count = bin(v).count('1')
            # Test method
            v_pop_count = bit_utils.popcount(v)

            self.assertEqual(v_pop_count, v_bin_count,
                             'popcount failed for integer %d' % v)



import collections
import heapq
import logging
import multiprocessing
import multiprocessing.queues
import multiprocessing.synchronize
import sys
import threading
import traceback

from six.moves import queue, range, zip, zip_longest

from smqtk.utils import SmqtkObject


def parallel_map(work_func, *sequences, **kwargs):
    """
    Generalized local parallelization helper for executing embarrassingly
    parallel functions on an iterable of input data. This function then yields
    work results for data, optionally in the order that they were provided to
    this function.

    By default, we act like ``itertools.izip`` in regards to input sequences,
    whereby we stop performing work as soon as one of the input sequences is
    exhausted. The optional keyword argument ``fill_void`` may be specified to
    enable sequence handling like ``itertools.zip_longest`` where the longest
    sequence determines what is iterated, and the value given to ``fill_void``
    is used as the fill value.

    This is intended to be able to replace ``multiprocessing.pool.Pool`` and
    ``multiprocessing.pool.ThreadPool`` uses with the added benefit of:
        - No set-up or clean-up needed
        - No performance loss compared to ``multiprocessing.pool`` classes
          for non-trivial work functions (like IO operations).
        - We iterate results as they are ready (optionally in order of
          input)
        - Lambda or on-the-fly function can be provided as the work function
          when using multiprocessing.
        - Buffered input/output queues so that mapping work of a very large
          input set doesn't overrun your memory (e.g. iterating over many large
          vectors/matrices).

    This function is, however, slower than multiprocessing.pool classes for
    trivial functions, like using the function ``ord`` over a set of
    characters.

    Input data given to ``sequences`` must be picklable in order to transport
    to worker threads/processes.

    :param work_func:
        Function that performs some work on input data, resulting in some
        returned value.

        When in multiprocessing mode, this cannot be a local function or a
        transport error will occur when trying to move the function to the
        worker process.
    :type work_func: (object, ...)-> object

    :param sequences: Input data to apply to the given ``work_func`` function.
        If more than one sequence is given, the function is called with an
        argument list consisting of the corresponding item of each sequence.
    :type sequences: collections.Iterable

    :param kwargs: Optionally available keyword arguments are as follows:

        - fill_void
            - Optional value that, if specified, activates sequence handling
              like that of ``__builtin__.map`` except that the value provided
              is used to

        - ordered
            - If results for input elements should be yielded in the same order
              as input elements. If False, we yield results as soon as they are
              collected.
            - type: bool
            - default: False

        - buffer_factor
            - Multiplier against the number of processes used to limit the
              growth size of the result queue coming from worker processes
              (``int(cores * buffer_factor)``). This is utilized so we don't
              overrun our RAM buffering results.
            - type: float
            - default: 2.0

        - cores
            - Optional specification of the number of threads/cores to use. If
              None, we will attempt to use all available threads/cores.
            - type: None | int | long
            - default: None

        - use_multiprocessing
            - Whether or not to use discrete processes as the parallelization
              agent vs python threads.
            - type: bool
            - default: False

        - heart_beat
            - Interval at which workers check for operational messages while
              waiting on locks (e.g. waiting to push or pull messages). This
              ensures that workers are not left hanging, or hang the program,
              when and error or interruption occurs, or when waiting on an full
              edge. This must be >0.
            - type: float
            - default: 0.001

        - name
            - Optional string name for identifying workers and logging
              messages. ``None`` means no names are added.
            - type: str
            - default: None

    :return: A new parallel results iterator that starts work on the input
        iterable when iterated.
    :rtype: ParallelResultsIterator


    Example
    -------
    >>> import math
    >>> result_iter = parallel_map(math.factorial, range(10), use_multiprocessing=True)
    >>> sorted(result_iter)
    [1, 1, 2, 6, 24, 120, 720, 5040, 40320, 362880]
    """
    # kwargs
    cores = kwargs.get('cores', None)
    ordered = kwargs.get('ordered', False)
    buffer_factor = kwargs.get('buffer_factor', 2.0)
    use_multiprocessing = kwargs.get('use_multiprocessing', False)
    heart_beat = kwargs.get('heart_beat', 0.001)
    fill_activate = 'fill_void' in kwargs
    fill_value = kwargs.get('fill_void', None)
    name = kwargs.get('name', None)

    if name:
        log = logging.getLogger(__name__ + '[%s]' % name)
    else:
        log = logging.getLogger(__name__)

    if heart_beat <= 0:
        raise ValueError("heart_beat must be >0.")

    if cores is None or cores <= 0:
        cores = multiprocessing.cpu_count()
        log.debug("Using all cores (%d)", cores)
    else:
        log.debug("Only using %d cores", cores)

    # Choose parallel types
    if use_multiprocessing:
        queue_t = multiprocessing.Queue
        worker_t = _WorkerProcess
    else:
        queue_t = queue.Queue
        worker_t = _WorkerThread

    queue_work = queue_t(int(cores * buffer_factor))
    queue_results = queue_t(int(cores * buffer_factor))

    log.log(1, "Constructing worker processes")
    workers = [worker_t(name, i, work_func, queue_work, queue_results,
                        heart_beat)
               for i in range(cores)]

    log.log(1, "Constructing feeder thread")
    feeder_thread = _FeedQueueThread(name, sequences, queue_work,
                                     len(workers), heart_beat, fill_activate,
                                     fill_value)

    return ParallelResultsIterator(name, ordered, use_multiprocessing,
                                   heart_beat, queue_work,
                                   queue_results, feeder_thread, workers)


class _TerminalPacket (object):
    """
    Signals a terminal message
    """


def is_terminal(p):
    """
    Check if a given packet is a terminal element.

    :param p: element to check
    :type p: object

    :return: If ``p`` is a terminal element
    :rtype: bool

    """
    return isinstance(p, _TerminalPacket)


class ParallelResultsIterator (SmqtkObject, collections.Iterator):

    def __init__(self, name, ordered, is_multiprocessing, heart_beat,
                 work_queue, results_queue,
                 feeder_thread, workers):
        """
        :type ordered: bool
        :type is_multiprocessing: bool
        :type heart_beat: float
        :type work_queue: Queue.Queue | multiprocessing.queues.Queue
        :type results_queue: Queue.Queue | multiprocessing.queues.Queue
        :type feeder_thread: _FeedQueueThread
        :type workers: list[_WorkerThread|_WorkerProcess]
        """
        if name:
            self.name = '[' + name + ']'
        else:
            self.name = ''

        self.ordered = ordered
        if self.ordered:
            self._log.debug("Maintaining result iteration order based on input "
                            "order")
        self.heart_beat = heart_beat
        self.is_multiprocessing = is_multiprocessing

        self.work_queue = work_queue
        self.results_queue = results_queue
        self.feeder_thread = feeder_thread
        self.workers = workers

        self.has_started_workers = False
        self.has_cleaned_up = False

        self.found_terminals = 0
        self.result_heap = []
        self.next_index = 0

        self.stop_event = threading.Event()
        self.stop_event_lock = threading.Lock()

    @property
    def _log(self):
        # Changing naming of logger returned from default
        return logging.getLogger(
            self.get_logger().name + self.name
        )

    def __repr__(self):
        sfx = ''
        if self.name:
            sfx = '[' + self.name + ']'
        return "<%(module)s.%(class)s%(sfx)s at %(address)s>" % {
            "module": self.__module__,
            "class": self.__class__.__name__,
            "sfx": sfx,
            "address": hex(id(self)),
        }

    def __next__(self):
        try:
            if not self.has_started_workers:
                self.start_workers()

            while (self.found_terminals < len(self.workers) and
                   not self.stopped()):
                packet = self.results_q_get()

                if is_terminal(packet):
                    self._log.log(1, 'Found terminal')
                    self.found_terminals += 1
                elif isinstance(packet[0], Exception):
                    ex, formatted_exc = packet
                    self._log.warn('Received exception: {}\n{}'.format(
                            ex, formatted_exc))
                    raise ex
                else:
                    i, result = packet
                    if self.ordered:
                        heapq.heappush(self.result_heap, (i, result))
                        if self.result_heap[0][0] == self.next_index:
                            _, result = heapq.heappop(self.result_heap)
                            self.next_index += 1
                            return result
                    else:
                        return result

            # Go through heap if there's anything in it
            if self.result_heap:
                _, result = heapq.heappop(self.result_heap)
                return result

            # Nothing left
            if not self.stopped():
                self._log.log(1, "Asserting empty queues on what looks like a "
                                 "full iteration.")
                self.assert_queues_empty()

            raise StopIteration()

        # If anything bad happens, stop iteration and workers.
        except Exception:
            self.stop()
            raise

    next = __next__

    def start_workers(self):
        """
        Start worker threads/processes
        """
        self._log.log(1, "Starting worker processes")
        for w in self.workers:
            w.start()

        self._log.log(1, "Starting feeder thread")
        # self.feeder_thread.daemon = True
        self.feeder_thread.start()

        self.has_started_workers = True

    def clean_up(self):
        """
        Clean up any live resources if we haven't done so already.
        """
        if self.has_started_workers and not self.has_cleaned_up:
            self._log.log(1, "Stopping feeder thread")
            self.feeder_thread.stop()
            self.feeder_thread.join()

            self._log.log(1, "Stopping workers")
            for w in self.workers:
                w.stop()
                w.join()

            if self.is_multiprocessing:
                self._log.log(1, "Closing/Joining process queues")
                for q in (self.work_queue, self.results_queue):
                    q.close()
                    q.join_thread()

            self.has_cleaned_up = True

    def stop(self):
        """
        Stop this iterator.

        This does not clean up resources (see ``clean_up`` for that).
        """
        with self.stop_event_lock:
            self.stop_event.set()
            self.clean_up()

    def stopped(self):
        """
        :return: if this iterator has been stopped
        :rtype: bool
        """
        return self.stop_event.is_set()

    def results_q_get(self):
        """
        Attempts to get something from the results queue.

        :raises StopIteration: when we've been told to stop.

        """
        while not self.stopped():
            try:
                return self.results_queue.get(timeout=self.heart_beat)
            except queue.Empty:
                pass
        raise StopIteration()

    def assert_queues_empty(self):
        # All work should be exhausted at this point
        if self.is_multiprocessing and sys.platform == 'darwin':
            # multiprocessing.Queue.qsize doesn't work on OSX
            # - Try to get something from each queue, expecting an empty
            #   exception.
            try:
                self.work_queue.get(block=False)
            except multiprocessing.queues.Empty:
                pass
            else:
                raise AssertionError("In queue not empty")
            try:
                self.results_queue.get(block=False)
            except multiprocessing.queues.Empty:
                pass
            else:
                raise AssertionError("Out queue not empty")
        else:
            assert self.work_queue.qsize() == 0, \
                "In queue not empty (%d)" % self.work_queue.qsize()
            assert self.results_queue.qsize() == 0, \
                "Out queue not empty (%d)" % self.results_queue.qsize()


class _FeedQueueThread (SmqtkObject, threading.Thread):
    """
    Helper thread for putting data into the work queue

    """

    def __init__(self, name, arg_sequences, q, num_terminal_packets, heart_beat,
                 do_fill, fill_value):
        threading.Thread.__init__(self, name=name)
        SmqtkObject.__init__(self)

        if name:
            self.name = '[' + name + ']'
        else:
            self.name = ''

        self.arg_sequences = arg_sequences
        self.q = q
        self.num_terminal_packets = num_terminal_packets
        self.heart_beat = heart_beat
        self.do_fill = do_fill
        self.fill_value = fill_value

        self._stop_event = threading.Event()

        # self.daemon = True

    @property
    def _log(self):
        # Changing naming of logger returned from default
        return logging.getLogger(
            self.get_logger().name + self.name
        )

    def stop(self):
        self._stop_event.set()

    def stopped(self):
        return self._stop_event.isSet()

    def run(self):
        self._log.log(1, "Starting")

        if self.do_fill:
            _zip = zip_longest
            _zip_kwds = {'fillvalue': self.fill_value}
        else:
            _zip = zip
            _zip_kwds = {}

        try:
            r = 0
            for args in _zip(*self.arg_sequences, **_zip_kwds):
                self.q_put((r, args))
                r += 1

                # If we're told to stop, immediately quit out of processing
                if self.stopped():
                    self._log.log(1, "Told to stop prematurely")
                    break
        except Exception as ex:
            self._log.warn("Caught exception %s", type(ex))
            self.q_put((ex, traceback.format_exc()))
            self.stop()
        else:
            self._log.log(1, "Sending in-queue terminal packets")
            for _ in range(self.num_terminal_packets):
                self.q_put(_TerminalPacket())
        finally:
            # Explicitly stop any nested parallel maps
            for s in self.arg_sequences:
                if isinstance(s, ParallelResultsIterator):
                    self._log.log(1, "Stopping nested parallel map: %s", s)
                    s.stop()

            self._log.log(1, "Closing")

    def q_put(self, val):
        """
        Try to put the given value into the output queue until it is inserted
        (if it was previously full), or the stop signal was given.

        :param val: value to put into the output queue.

        """
        put = False
        while not put and not self.stopped():
            try:
                self.q.put(val, timeout=self.heart_beat)
                put = True
            except queue.Full:
                pass


class _Worker (SmqtkObject):

    def __init__(self, name, i, work_function, in_q, out_q, heart_beat):
        """
        :type name: str
        :type i: int
        :type work_function: (*args) -> object
        :type in_q: multiprocessing.queues.Queue
        :type out_q: multiprocessing.queues.Queue
        :type heart_beat: float
        """
        if name:
            self.name = '[' + name + '::%d]' % i
        else:
            self.name = '::%d' % i

        self.i = i
        self.work_function = work_function
        self.in_q = in_q
        self.out_q = out_q
        self.heart_beat = heart_beat
        self._log.log(1, "Making process worker (%d, %s, %s)", i, in_q, out_q)

        self._stop_event = self._make_event()

    @property
    def _log(self):
        return logging.getLogger(
            self.get_logger().name + self.name
        )

    def _make_event(self):
        raise NotImplementedError()

    def stop(self):
        self._stop_event.set()

    def stopped(self):
        return self._stop_event.is_set()

    def run(self):
        try:
            packet = self.q_get()
            while not self.stopped():
                if is_terminal(packet):
                    self._log.log(1, "sending terminal")
                    self.q_put(packet)
                    self.stop()
                elif isinstance(packet[0], Exception):
                    # Pass exception along
                    self.q_put(packet)
                    self.stop()
                else:
                    i, args = packet
                    result = self.work_function(*args)
                    self.q_put((i, result))
                    packet = self.q_get()
        # Transport back any exceptions raised
        except Exception as ex:
            self._log.warn("Caught exception %s", type(ex))
            self.q_put((ex, traceback.format_exc()))
            self.stop()
        finally:
            self._log.log(1, "Closing")

    def q_get(self):
        """
        Try to get a value from the queue while keeping an eye out for an exit
        request.

        :return: next value on the input queue

        """
        while not self.stopped():
            try:
                return self.in_q.get(timeout=self.heart_beat)
            except queue.Empty:
                pass

    def q_put(self, val):
        """
        Try to put the given value into the output queue while keeping an eye
        out for an exit request.

        :param val: value to put into the output queue.

        """
        put = False
        while not put and not self.stopped():
            try:
                self.out_q.put(val, timeout=self.heart_beat)
                put = True
            except queue.Full:
                pass


class _WorkerProcess (_Worker, multiprocessing.Process):

    def __init__(self, name, i, work_function, in_q, out_q, heart_beat):
        multiprocessing.Process.__init__(self)
        _Worker.__init__(self, name, i, work_function, in_q, out_q, heart_beat)
        # if name:
        #     self.name = name + "::%d" % i

        # self.daemon = True

    def _make_event(self):
        return multiprocessing.Event()

    run = _Worker.run


class _WorkerThread (_Worker, threading.Thread):

    def __init__(self, name, i, work_function, in_q, out_q, heart_beat):
        threading.Thread.__init__(self)
        _Worker.__init__(self, name, i, work_function, in_q, out_q, heart_beat)
        # if name:
        #     self.name = name + "::%d" % i

        # self.daemon = True

    def _make_event(self):
        return threading.Event()

    run = _Worker.run



from collections import deque
import io
import itertools
import logging
import multiprocessing
import multiprocessing.pool

import numpy
import PIL.Image
import PIL.ImageFile
import six
# noinspection PyUnresolvedReferences
from six.moves import range, zip

from smqtk.algorithms.descriptor_generator import \
    DescriptorGenerator, \
    DFLT_DESCRIPTOR_FACTORY
from smqtk.representation.data_element import from_uri
from smqtk.utils.bin_utils import report_progress

try:
    import caffe
except ImportError as ex:
    logging.getLogger(__name__).warning("Failed to import caffe module: %s",
                                        str(ex))
    caffe = None


__author__ = 'paul.tunison@kitware.com, jacob.becker@kitware.com'

__all__ = [
    "CaffeDescriptorGenerator",
]


class CaffeDescriptorGenerator (DescriptorGenerator):
    """
    Compute images against a Caffe model, extracting a layer as the content
    descriptor.
    """

    @classmethod
    def is_usable(cls):
        valid = caffe is not None
        if not valid:
            cls.get_logger().debug("Caffe python module cannot be imported")
        return valid

    def __init__(self, network_prototxt_uri, network_model_uri, image_mean_uri,
                 return_layer='fc7',
                 batch_size=1, use_gpu=False, gpu_device_id=0,
                 network_is_bgr=True, data_layer='data',
                 load_truncated_images=False, pixel_rescale=None,
                 input_scale=None):
        """
        Create a Caffe CNN descriptor generator

        :param network_prototxt_uri: URI to the text file defining the
            network layout.
        :type network_prototxt_uri: str

        :param network_model_uri: URI to the trained ``.caffemodel``
            file to use.
        :type network_model_uri: str

        :param image_mean_uri: URI to the image mean ``.binaryproto`` or
            ``.npy`` file.
        :type image_mean_uri: str | file | StringIO.StringIO

        :param return_layer: The label of the layer we take data from to compose
            output descriptor vector.
        :type return_layer: str

        :param batch_size: The maximum number of images to process in one feed
            forward of the network. This is especially important for GPUs since
            they can only process a batch that will fit in the GPU memory space.
        :type batch_size: int

        :param use_gpu: If Caffe should try to use the GPU
        :type use_gpu: bool

        :param gpu_device_id: Integer ID of the GPU device to use. Only used if
            ``use_gpu`` is True.
        :type gpu_device_id: int

        :param network_is_bgr: If the network is expecting BGR format pixels.
            For example, the BVLC default caffenet does (thus the default is
            True).
        :type network_is_bgr: bool

        :param data_layer: String label of the network's data layer.
            We assume its 'data' by default.
        :type data_layer: str

        :param load_truncated_images: If we should be lenient and force loading
            of truncated image bytes. This is False by default.
        :type load_truncated_images: bool

        :param pixel_rescale: Re-scale image pixel values before being
            transformed by caffe (before mean subtraction, etc)
            into the given tuple ``(min, max)`` range. By default, images are
            loaded in the ``[0, 255]`` range. Refer to the image mean being used
            for desired input pixel scale.
        :type pixel_rescale: None | (float, float)

        :param input_scale: Optional floating-point scalar value to scale values
            of caffe network input data AFTER mean subtraction. This value is
            directly multiplied against the pixel values.
        :type input_scale: None | float

        """
        super(CaffeDescriptorGenerator, self).__init__()

        self.network_prototxt_uri = str(network_prototxt_uri)
        self.network_model_uri = str(network_model_uri)
        self.image_mean_uri = image_mean_uri

        self.return_layer = str(return_layer)
        self.batch_size = int(batch_size)

        self.use_gpu = bool(use_gpu)
        self.gpu_device_id = int(gpu_device_id)

        self.network_is_bgr = bool(network_is_bgr)
        self.data_layer = str(data_layer)

        self.load_truncated_images = bool(load_truncated_images)
        self.pixel_rescale = pixel_rescale
        self.input_scale = input_scale

        assert self.batch_size > 0, \
            "Batch size must be greater than 0 (got %d)" \
            % self.batch_size
        assert self.gpu_device_id >= 0, \
            "GPU Device ID must be greater than 0 (got %d)" \
            % self. gpu_device_id

        # Network setup variables
        self.network = None
        self.net_data_shape = ()
        self.transformer = None

        self._setup_network()

    def __getstate__(self):
        return self.get_config()

    def __setstate__(self, state):
        # This works because configuration parameters exactly match up with
        # instance attributes.
        self.__dict__.update(state)
        self._setup_network()

    def _set_caffe_mode(self):
        """
        Set the appropriate Caffe mode on the current thread/process.
        """
        if self.use_gpu:
            self._log.debug("Using GPU")
            caffe.set_device(self.gpu_device_id)
            caffe.set_mode_gpu()
        else:
            self._log.debug("using CPU")
            caffe.set_mode_cpu()

    def _setup_network(self):
        """
        Initialize Caffe and the network
        """
        self._set_caffe_mode()

        # Questions:
        #   - ``caffe.TEST`` indicates phase of either TRAIN or TEST
        self._log.debug("Initializing network")
        network_prototxt_element = from_uri(self.network_prototxt_uri)
        network_model_element = from_uri(self.network_model_uri)
        self._log.debug("Loading Caffe network from network/model configs")
        self.network = caffe.Net(network_prototxt_element.write_temp(),
                                 caffe.TEST,
                                 weights=network_model_element.write_temp())
        network_prototxt_element.clean_temp()
        network_model_element.clean_temp()
        # Assuming the network has a 'data' layer and notion of data shape
        self.net_data_shape = self.network.blobs[self.data_layer].data.shape
        self._log.debug("Network data shape: %s", self.net_data_shape)

        # Crating input data transformer
        self._log.debug("Initializing data transformer")
        self.transformer = caffe.io.Transformer(
            {self.data_layer: self.network.blobs[self.data_layer].data.shape}
        )
        self._log.debug("Initializing data transformer -> %s",
                        self.transformer.inputs)

        self._log.debug("Loading image mean")
        image_mean_elem = from_uri(self.image_mean_uri)
        image_mean_bytes = image_mean_elem.get_bytes()
        try:
            a = numpy.load(io.BytesIO(image_mean_bytes))
            self._log.info("Loaded image mean from numpy bytes")
        except IOError:
            self._log.debug("Image mean file not a numpy array, assuming "
                            "URI to protobuf binary.")
            # noinspection PyUnresolvedReferences
            blob = caffe.proto.caffe_pb2.BlobProto()
            blob.ParseFromString(image_mean_bytes)
            a = numpy.array(caffe.io.blobproto_to_array(blob))
            assert a.shape[0] == 1, \
                "Input image mean blob protobuf consisted of more than one " \
                "image. Not sure how to handle this yet."
            a = a.reshape(a.shape[1:])
            self._log.info("Loaded image mean from protobuf bytes")
        assert a.shape[0] in [1, 3], \
            "Currently asserting that we either get 1 or 3 channel images. " \
            "Got a %d channel image." % a[0]
        # TODO: Instead of always using pixel mean, try to use image-mean if
        #       given. Might have to rescale if image/data layer shape is
        #       different.
        a_mean = a.mean(1).mean(1)
        self._log.debug("Initializing data transformer -- mean")
        self.transformer.set_mean(self.data_layer, a_mean)

        self._log.debug("Initializing data transformer -- transpose")
        self.transformer.set_transpose(self.data_layer, (2, 0, 1))
        if self.network_is_bgr:
            self._log.debug("Initializing data transformer -- channel swap")
            self.transformer.set_channel_swap(self.data_layer, (2, 1, 0))
        if self.input_scale:
            self._log.debug("Initializing data transformer -- input scale")
            self.transformer.set_input_scale(self.data_layer, self.input_scale)

    def get_config(self):
        """
        Return a JSON-compliant dictionary that could be passed to this class's
        ``from_config`` method to produce an instance with identical
        configuration.

        In the common case, this involves naming the keys of the dictionary
        based on the initialization argument names as if it were to be passed
        to the constructor via dictionary expansion.

        :return: JSON type compliant configuration dictionary.
        :rtype: dict

        """
        return {
            "network_prototxt_uri": self.network_prototxt_uri,
            "network_model_uri": self.network_model_uri,
            "image_mean_uri": self.image_mean_uri,
            "return_layer": self.return_layer,
            "batch_size": self.batch_size,
            "use_gpu": self.use_gpu,
            "gpu_device_id": self.gpu_device_id,
            "network_is_bgr": self.network_is_bgr,
            "data_layer": self.data_layer,
            "load_truncated_images": self.load_truncated_images,
            "pixel_rescale": self.pixel_rescale,
            "input_scale": self.input_scale,
        }

    def valid_content_types(self):
        """
        :return: A set valid MIME type content types that this descriptor can
            handle.
        :rtype: set[str]
        """
        return {
            'image/tiff',
            'image/png',
            'image/jpeg',
        }

    def _compute_descriptor(self, data):
        raise NotImplementedError("Shouldn't get here as "
                                  "compute_descriptor[_async] is being "
                                  "overridden")

    def compute_descriptor(self, data, descr_factory=DFLT_DESCRIPTOR_FACTORY,
                           overwrite=False):
        """
        Given some data, return a descriptor element containing a descriptor
        vector.

        :raises RuntimeError: Descriptor extraction failure of some kind.
        :raises ValueError: Given data element content was not of a valid type
            with respect to this descriptor.

        :param data: Some kind of input data for the feature descriptor.
        :type data: smqtk.representation.DataElement

        :param descr_factory: Factory instance to produce the wrapping
            descriptor element instance. The default factory produces
            ``DescriptorMemoryElement`` instances by default.
        :type descr_factory: smqtk.representation.DescriptorElementFactory

        :param overwrite: Whether or not to force re-computation of a descriptor
            vector for the given data even when there exists a precomputed
            vector in the generated DescriptorElement as generated from the
            provided factory. This will overwrite the persistently stored vector
            if the provided factory produces a DescriptorElement implementation
            with such storage.
        :type overwrite: bool

        :return: Result descriptor element. UUID of this output descriptor is
            the same as the UUID of the input data element.
        :rtype: smqtk.representation.DescriptorElement

        """
        m = self.compute_descriptor_async([data], descr_factory, overwrite,
                                          procs=1)
        return m[data.uuid()]

    def compute_descriptor_async(self, data_iter,
                                 descr_factory=DFLT_DESCRIPTOR_FACTORY,
                                 overwrite=False, procs=None, **kwds):
        """
        Asynchronously compute feature data for multiple data items.

        :param data_iter: Iterable of data elements to compute features for.
            These must have UIDs assigned for feature association in return
            value.
        :type data_iter: collections.Iterable[smqtk.representation.DataElement]

        :param descr_factory: Factory instance to produce the wrapping
            descriptor element instance. The default factory produces
            ``DescriptorMemoryElement`` instances by default.
        :type descr_factory: smqtk.representation.DescriptorElementFactory

        :param overwrite: Whether or not to force re-computation of a descriptor
            vectors for the given data even when there exists precomputed
            vectors in the generated DescriptorElements as generated from the
            provided factory. This will overwrite the persistently stored
            vectors if the provided factory produces a DescriptorElement
            implementation such storage.
        :type overwrite: bool

        :param procs: Optional specification of how many processors to use
            when pooling sub-tasks. If None, we attempt to use all available
            cores.
        :type procs: int | None

        :raises ValueError: An input DataElement was of a content type that we
            cannot handle.

        :return: Mapping of input DataElement UUIDs to the computed descriptor
            element for that data. DescriptorElement UUID's are congruent with
            the UUID of the data element it is the descriptor of.
        :rtype: dict[collections.Hashable,
                     smqtk.representation.DescriptorElement]

        """
        self._set_caffe_mode()

        # Create DescriptorElement instances for each data elem.
        #: :type: dict[collections.Hashable, smqtk.representation.DataElement]
        data_elements = {}
        #: :type: dict[collections.Hashable, smqtk.representation.DescriptorElement]
        descr_elements = {}
        self._log.debug("Checking content types; aggregating data/descriptor "
                        "elements.")
        prog_rep_state = [0] * 7
        for data in data_iter:
            ct = data.content_type()
            if ct not in self.valid_content_types():
                self._log.error("Cannot compute descriptor from content type "
                                "'%s' data: %s)" % (ct, data))
                raise ValueError("Cannot compute descriptor from content type "
                                 "'%s' data: %s)" % (ct, data))
            data_elements[data.uuid()] = data
            descr_elements[data.uuid()] = \
                descr_factory.new_descriptor(self.name, data.uuid())
            report_progress(self._log.debug, prog_rep_state, 1.0)
        self._log.debug("Given %d unique data elements", len(data_elements))

        # Reduce procs down to the number of elements to process if its smaller
        if len(data_elements) < (procs or multiprocessing.cpu_count()):
            procs = len(data_elements)
        if procs == 0:
            raise ValueError("No data elements provided")

        # For thread safely, only use .append() and .popleft() (queue)
        uuid4proc = deque()

        def check_get_uuid(descriptor_elem):
            if overwrite or not descriptor_elem.has_vector():
                # noinspection PyUnresolvedReferences
                uuid4proc.append(descriptor_elem.uuid())

        # Using thread-pool due to in-line function + updating local deque
        p = multiprocessing.pool.ThreadPool(procs)
        try:
            p.map(check_get_uuid, six.itervalues(descr_elements))
        finally:
            p.close()
            p.join()
        del p
        self._log.debug("%d descriptors already computed",
                        len(data_elements) - len(uuid4proc))

        if uuid4proc:
            self._log.debug("Converting deque to tuple for segmentation")
            uuid4proc = tuple(uuid4proc)

            # Split UUIDs into groups equal to our batch size, and an option
            # tail group that is less than our batch size.
            tail_size = len(uuid4proc) % self.batch_size
            batch_groups = (len(uuid4proc) - tail_size) // self.batch_size
            self._log.debug("Processing %d batches of size %d", batch_groups,
                            self.batch_size)
            if tail_size:
                self._log.debug("Processing tail group of size %d", tail_size)

            if batch_groups:
                for g in range(batch_groups):
                    self._log.debug("Starting batch: %d of %d",
                                    g + 1, batch_groups)
                    batch_uuids = \
                        uuid4proc[g * self.batch_size:(g + 1) * self.batch_size]
                    self._process_batch(batch_uuids, data_elements,
                                        descr_elements, procs,
                                        kwds.get('use_mp', True))

            if tail_size:
                batch_uuids = uuid4proc[-tail_size:]
                self._log.debug("Starting tail batch (size=%d)",
                                len(batch_uuids))
                self._process_batch(batch_uuids, data_elements, descr_elements,
                                    procs, kwds.get('use_mp', True))

        self._log.debug("forming output dict")
        return dict((data_elements[k].uuid(), descr_elements[k])
                    for k in data_elements)

    def _process_batch(self, uuids4proc, data_elements, descr_elements, procs,
                       use_mp):
        """
        Run a number of data elements through the network, based on the number
        of UUIDs given, returning the vectors of

        :param uuids4proc: UUIDs of the source data to run in the network as a
            batch.
        :type uuids4proc: collections.Sequence[collections.Hashable]

        :param data_elements: Mapping of UUID to data element for input data.
        :type data_elements: dict[collections.Hashable,
                                  smqtk.representation.DataElement]

        :param descr_elements: Mapping of UUID to descriptor element based on
            input data elements.
        :type descr_elements: dict[collections.Hashable,
                                   smqtk.representation.DescriptorElement]

        :param procs: The number of asynchronous processes to run for loading
            images. This may be None to just use all available cores.
        :type procs: None | int

        :param use_mp: Whether or not to use a multiprocessing pool or a thread
            pool.
        :type use_mp: bool

        """
        self._log.debug("Updating network data layer shape (%d images)",
                        len(uuids4proc))
        self.network.blobs[self.data_layer].reshape(len(uuids4proc),
                                                    *self.net_data_shape[1:4])

        self._log.debug("Loading image pixel arrays")
        uid_num = len(uuids4proc)

        if use_mp:
            p = multiprocessing.Pool(procs)
        else:
            p = multiprocessing.pool.ThreadPool(procs)

        img_arrays = p.map(
            _process_load_img_array,
            list(zip(
                (data_elements[uid] for uid in uuids4proc),
                itertools.repeat(self.transformer, uid_num),
                itertools.repeat(self.data_layer, uid_num),
                itertools.repeat(self.load_truncated_images, uid_num),
                itertools.repeat(self.pixel_rescale, uid_num),
            ))
        )
        p.close()
        p.join()

        self._log.debug("Loading image bytes into network layer '%s'",
                        self.data_layer)
        self.network.blobs[self.data_layer].data[...] = img_arrays

        self._log.debug("Moving network forward")
        self.network.forward()
        descriptor_list = self.network.blobs[self.return_layer].data

        self._log.debug("extracting return layer '%s' into descriptors",
                        self.return_layer)
        for uid, v in zip(uuids4proc, descriptor_list):
            if v.ndim > 1:
                # In case caffe generates multidimensional array (rows, 1, 1)
                descr_elements[uid].set_vector(numpy.ravel(v))
            else:
                descr_elements[uid].set_vector(v)


def _process_load_img_array(input_tuple):
    """
    Helper function for multiprocessing image data loading

    :param data_element: DataElement providing the bytes
    :type data_element: smqtk.representation.DataElement

    :param transformer: Caffe Transformer instance for pre-processing
    :type transformer: caffe.io.Transformer

    :param load_truncated_images: If PIL should be allowed to load truncated
        image data. If false, and exception will be raised when encountering
        such imagery.

    :return: Pre-processed numpy array.

    """
    (data_element, transformer, data_layer, load_truncated_images,
     pixel_rescale) = input_tuple
    PIL.ImageFile.LOAD_TRUNCATED_IMAGES = load_truncated_images
    img = PIL.Image.open(io.BytesIO(data_element.get_bytes()))
    if img.mode != "RGB":
        img = img.convert("RGB")
    # Caffe natively uses float types (32-bit)
    try:
        # This can fail if the image is truncated and we're not allowing the
        # loading of those images
        img_a = numpy.asarray(img, numpy.float32)
    except Exception:
        logging.getLogger(__name__).error(
            "Failed array-ifying data element. Image may be truncated: %s",
            data_element
        )
        raise
    assert img_a.ndim == 3, \
        "Loaded invalid RGB image with shape %s" \
        % img_a.shape
    if pixel_rescale:
        pmin, pmax = min(pixel_rescale), max(pixel_rescale)
        r = pmax - pmin
        img_a = (img_a / (255. / r)) + pmin
    img_at = transformer.preprocess(data_layer, img_a)
    return img_at



import threading

from six.moves import cPickle as pickle

from smqtk.exceptions import ReadOnlyError
from smqtk.representation import DataElement, DataSet, get_data_element_impls
from smqtk.utils import merge_dict, plugin, SimpleTimer


class DataMemorySet (DataSet):
    """
    In-memory DataSet implementation.

    This implementation maintains an in-memory mapping of stored DataElement
    original UUID to the DataElement instance.

    An optional writable DataElement may be provided to which the current set's
    map state is cached. This cache is updated every time new data elements are
    added to this set..

    """

    @classmethod
    def is_usable(cls):
        """
        Check whether this data set implementations is available for use.

        This is always true for this implementation as there are no required 3rd
        party dependencies

        :return: Boolean determination of whether this implementation is usable.
        :rtype: bool

        """
        return True

    @classmethod
    def get_default_config(cls):
        """
        Generate and return a default configuration dictionary for this class.
        This will be primarily used for generating what the configuration
        dictionary would look like for this class without instantiating it.

        By default, we observe what this class's constructor takes as arguments,
        turning those argument names into configuration dictionary keys. If any
        of those arguments have defaults, we will add those values into the
        configuration dictionary appropriately. The dictionary returned should
        only contain JSON compliant value types.

        It is not be guaranteed that the configuration dictionary returned
        from this method is valid for construction of an instance of this class.

        :return: Default configuration dictionary for the class.
        :rtype: dict

        """
        c = super(DataMemorySet, cls).get_default_config()
        c['cache_element'] = plugin.make_config(get_data_element_impls())
        return c

    @classmethod
    def from_config(cls, c, merge_default=True):
        """
        Instantiate a new instance of this class given the configuration
        JSON-compliant dictionary encapsulating initialization arguments.

        This method should not be called via super unless an instance of the
        class is desired.

        :param c: JSON compliant dictionary encapsulating
            a configuration.
        :type c: dict

        :param merge_default: Merge the given configuration on top of the
            default provided by ``get_default_config``.
        :type merge_default: bool

        :return: Constructed instance from the provided config.
        :rtype: DataMemorySet

        """
        if merge_default:
            c = merge_dict(cls.get_default_config(), c)

        cache_element = None
        if c['cache_element'] and c['cache_element']['type']:
            cache_element = plugin.from_plugin_config(c['cache_element'],
                                                      get_data_element_impls())
        c['cache_element'] = cache_element

        return super(DataMemorySet, cls).from_config(c, False)

    def __init__(self, cache_element=None, pickle_protocol=-1):
        """
        Initialize a new in-memory data set instance.

        :param cache_element: Optional data element to store/load a cache of
            this data set's contents into. Cache loading, if the element has
            bytes, will occur in this constructor. Cache writing will only occur
            after adding one or more elements.

            This can be optionally turned on after creating/using this data set
            for a while by setting a valid element to the ``cache_element``
            attribute and calling the ``.cache()`` method. When
            ``cache_element`` is not set, the ``cache()`` method does nothing.
        :type cache_element: None | smqtk.representation.DataElement

        :param pickle_protocol: Pickling protocol to use. We will use -1 by
            default (latest version, probably binary).
        :type pickle_protocol: int

        """
        super(DataMemorySet, self).__init__()

        # Mapping of UUIDs to DataElement instances
        #: :type: dict[collections.Hashable, DataElement]
        self._element_map = {}
        self._element_map_lock = threading.RLock()

        # Optional path to a file that will act as a cache of our internal
        # table
        self.cache_element = cache_element
        if cache_element and not cache_element.is_empty():
            #: :type: dict[collections.Hashable, DataElement]
            self._element_map = pickle.loads(cache_element.get_bytes())

        self.pickle_protocol = pickle_protocol

    def __iter__(self):
        """
        :return: Generator over the DataElements contained in this set in no
            particular order.
        """
        # making copy of UUIDs so we don't block when between yields, as well
        # as so we aren't walking a possibly modified map
        uuids = self.uuids()
        with self._element_map_lock:
            for k in uuids:
                yield self._element_map[k]

    def cache(self):
        """
        Cache the current table if a cache has been configured.
        """
        if self.cache_element:
            if self.cache_element.is_read_only():
                raise ReadOnlyError("Cache element (%s) is read-only."
                                    % self.cache_element)

            with self._element_map_lock:
                with SimpleTimer("Caching memory data-set table",
                                 self._log.debug):
                    self.cache_element.set_bytes(
                        pickle.dumps(self._element_map, self.pickle_protocol)
                    )

    def get_config(self):
        """
        This implementation has no configuration properties.

        :return: JSON type compliant configuration dictionary.
        :rtype: dict

        """
        c = merge_dict(self.get_default_config(), {
            "pickle_protocol": self.pickle_protocol,
        })
        if self.cache_element:
            c['cache_element'] = merge_dict(
                c['cache_element'],
                plugin.to_plugin_config(self.cache_element)
            )
        return c

    def count(self):
        """
        :return: The number of data elements in this set.
        :rtype: int
        """
        with self._element_map_lock:
            return len(self._element_map)

    def uuids(self):
        """
        :return: A new set of uuids represented in this data set.
        :rtype: set
        """
        with self._element_map_lock:
            return set(self._element_map)

    def has_uuid(self, uuid):
        """
        Test if the given uuid refers to an element in this data set.

        :param uuid: Unique ID to test for inclusion. This should match the
            type that the set implementation expects or cares about.

        :return: True if the given uuid matches an element in this set, or
            False if it does not.
        :rtype: bool

        """
        with self._element_map_lock:
            return uuid in self._element_map

    def add_data(self, *elems):
        """
        Add the given data element(s) instance to this data set.

        :param elems: Data element(s) to add
        :type elems: smqtk.representation.DataElement

        """
        with self._element_map_lock:
            added_elements = False
            for e in elems:
                assert isinstance(e, DataElement), \
                    "Expected DataElement instance, got '%s' instance instead" \
                    % type(e)
                self._element_map[e.uuid()] = e
                added_elements = True
            if added_elements:
                self.cache()

    def get_data(self, uuid):
        """
        Get the data element the given uuid references, or raise an
        exception if the uuid does not reference any element in this set.

        :raises KeyError: If the given uuid does not refer to an element in
            this data set.

        :param uuid: The uuid of the element to retrieve.

        :return: The data element instance for the given uuid.
        :rtype: smqtk.representation.DataElement

        """
        with self._element_map_lock:
            return self._element_map[uuid]


DATA_SET_CLASS = DataMemorySet



from smqtk.representation import (
    SmqtkRepresentation,
    get_classification_element_impls
)
from smqtk.utils.plugin import make_config
from smqtk.utils import merge_dict


__author__ = "paul.tunison@kitware.com"


class ClassificationElementFactory (SmqtkRepresentation):
    """
    Factory class for producing ClassificationElement instances of a specified
    type and configuration.
    """

    # noinspection PyShadowingBuiltins
    def __init__(self, type, type_config):
        """
        Initialize the factory to produce ClassificationElement instances of the
        given type from the given configuration.

        :param type: Python implementation type of the ClassifierElement to
            produce
        :type type: type

        :param type_config: Configuration dictionary that will be passed
            ``from_config`` class method of given ``type``.
        :type type_config: dict

        """
        #: :type: type | smqtk.representation.ClassificationElement
        self.type = type
        self.type_config = type_config

    #
    # Class methods
    #

    @classmethod
    def get_default_config(cls):
        """
        Generate and return a default configuration dictionary for this class.
        This will be primarily used for generating what the configuration
        dictionary would look like for this class without instantiating it.

        It is not be guaranteed that the configuration dictionary returned
        from this method is valid for construction of an instance of this class.

        :return: Default configuration dictionary for the class.
        :rtype: dict

        """
        return make_config(get_classification_element_impls())

    @classmethod
    def from_config(cls, config_dict, merge_default=True):
        """
        Instantiate a new instance of this class given the configuration
        JSON-compliant dictionary encapsulating initialization arguments.

        This method should not be called via super unless and instance of the
        class is desired.

        :param config_dict: JSON compliant dictionary encapsulating
            a configuration.
        :type config_dict: dict

        :param merge_default: Merge the given configuration on top of the
            default provided by ``get_default_config``.
        :type merge_default: bool

        :return: Constructed instance from the provided config.
        :rtype: ClassificationElementFactory

        """
        if merge_default:
            mc = cls.get_default_config()
            merge_dict(mc, config_dict)
            config_dict = mc

        return ClassificationElementFactory(
            get_classification_element_impls()[config_dict['type']],
            config_dict[config_dict['type']]
        )

    def get_config(self):
        type_name = self.type.__name__
        return {
            "type": type_name,
            type_name: self.type_config,
        }

    # noinspection PyShadowingBuiltins
    def new_classification(self, type, uuid):
        """
        Create a new ClassificationElement instance of the configured
        implementation.

        :param type: Type of classifier. This is usually the name of the
            classifier that generated this result.
        :type type: str

        :param uuid: UUID to associate with the classification.
        :type uuid: collections.Hashable

        :return: New ClassificationElement instance.
        :rtype: smqtk.representation.ClassificationElement

        """
        return self.type.from_config(self.type_config, type, uuid)

    # noinspection PyShadowingBuiltins
    def __call__(self, type, uuid):
        """
        Create a new ClassificationElement instance of the configured
        implementation.

        :param type: Type of classifier. This is usually the name of the
            classifier that generated this result.
        :type type: str

        :param uuid: UUID to associate with the classification.
        :type uuid: collections.Hashable

        :return: New ClassificationElement instance.
        :rtype: smqtk.representation.ClassificationElement

        """
        return self.new_classification(type, uuid)



import os.path
from unittest import TestCase

from smqtk.tests import TEST_DATA_DIR
from smqtk.utils.file_utils import file_mimetype_tika

try:
    from tika import detector as tika_detector
except ImportError:
    tika_detector = None


if tika_detector is not None:

    class TestFile_mimetype_tika(TestCase):

        def test_file_doesnt_exist(self):
            try:
                file_mimetype_tika('/this/path/probably/doesnt/exist.txt')
            except IOError as ex:
                self.assertEqual(ex.errno, 2,
                                 "Expected directory IO error #2. "
                                 "Got %d" % ex.errno)

        def test_directory_provided(self):
            try:
                file_mimetype_tika(TEST_DATA_DIR)
            except IOError as ex:
                self.assertEqual(ex.errno, 21,
                                 "Expected directory IO error #21. "
                                 "Got %d" % ex.errno)

        def test_get_mimetype_lenna(self):
            m = file_mimetype_tika(
                os.path.join(TEST_DATA_DIR, 'Lenna.png')
            )
            self.assertEqual(m, 'image/png')

        def test_get_mimetype_no_extension(self):
            m = file_mimetype_tika(
                os.path.join(TEST_DATA_DIR, 'text_file')
            )
            self.assertEqual(m, 'text/plain')



import errno
import mock
import os
import unittest

from smqtk.utils import file_utils


class TestSafeCreateDir (unittest.TestCase):

    @mock.patch('smqtk.utils.file_utils.os.makedirs')
    def test_noExists(self, mock_os_makedirs):
        dir_path = "/some/directory/somewhere"
        p = file_utils.safe_create_dir(dir_path)

        self.assertTrue(mock_os_makedirs.called)
        self.assertEquals(p, dir_path)

    @mock.patch('smqtk.utils.file_utils.os.path.exists')
    @mock.patch('smqtk.utils.file_utils.os.makedirs')
    def test_existError_alreadyExists(self, mock_os_makedirs, mock_osp_exists):
        mock_os_makedirs.side_effect = OSError(errno.EEXIST,
                                               "Existing directory")

        mock_osp_exists.return_value = True

        dir_path = '/existing/dir'
        p = file_utils.safe_create_dir(dir_path)

        self.assertTrue(mock_os_makedirs.called)
        self.assertTrue(mock_osp_exists.called)
        mock_osp_exists.assert_called_once_with(dir_path)
        self.assertEqual(p, dir_path)

    @mock.patch('smqtk.utils.file_utils.os.path.exists')
    @mock.patch('smqtk.utils.file_utils.os.makedirs')
    def test_existError_noExist(self, mock_os_makedirs, mock_osp_exists):
        mock_os_makedirs.side_effect = OSError(errno.EEXIST,
                                               "Existing directory")
        mock_osp_exists.return_value = False

        dir_path = '/some/dir'
        self.assertRaises(OSError, file_utils.safe_create_dir, dir_path)

        mock_os_makedirs.assert_called_once_with(dir_path)
        mock_osp_exists.assert_called_once_with(dir_path)

    @mock.patch('smqtk.utils.file_utils.os.path.exists')
    @mock.patch('smqtk.utils.file_utils.os.makedirs')
    def test_otherOsError(self, mock_os_makedirs, mock_osp_exists):
        mock_os_makedirs.side_effect = OSError(errno.EACCES,
                                               "Permission Denied")

        dir_path = '/some/dir'
        self.assertRaises(OSError, file_utils.safe_create_dir, dir_path)

        mock_os_makedirs.assert_called_once_with(dir_path)
        self.assertFalse(mock_osp_exists.called)

    @mock.patch('smqtk.utils.file_utils.os.makedirs')
    def test_otherException(self, mock_os_makedirs):
        mock_os_makedirs.side_effect = RuntimeError("Some other exception")

        dir_path = 'something'
        self.assertRaises(RuntimeError, file_utils.safe_create_dir, dir_path)

        mock_os_makedirs.assert_called_once_with(os.path.abspath(dir_path))
